import os
import time
import warnings
import random
from itertools import chain, combinations
from typing import Tuple, Any, Dict, Type, Optional, List, Union, Sequence
import string

from typing_extensions import Literal
import torch
import optuna
from optuna.trial import TrialState
import pandas as pd
import numpy as np
from torch.utils.tensorboard.writer import SummaryWriter
from torch.utils.tensorboard.summary import hparams
import plotly.express as px
from optuna.pruners import MedianPruner, BasePruner
from optuna.trial._base import BaseTrial  # noqa
from optuna.study import Study
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler  # noqa

from gdeep.utility import _inner_refactor_scalars, KnownWarningSilencer  # noqa
from gdeep.trainer import Trainer
from gdeep.search import Benchmark, _benchmarking_param
from gdeep.visualization import plotly2tensor
from ..utility import save_model_and_optimizer
from .hpo_config import HPOConfig

from gdeep.utility.custom_types import Tensor, Array

SEARCH_METRICS = ("loss", "accuracy")


class GiottoSummaryWriter(SummaryWriter):
    def add_hparams(
        self,
        hparam_dict: Dict[str, Any],
        metric_dict: Dict[str, Any],
        hparam_domain_discrete: Optional[Dict[str, List[Any]]] = None,
        run_name: Optional[str] = None,
        scalars_lists: Optional[List[List[Tuple[Any, int]]]] = None,
        best_not_last: bool = False,
    ):
        """Add a set of hyperparameters to be compared in TensorBoard.
        Args:
            hparam_dict (dict):
                Each key-value pair in the dictionary is the
                name of the hyper parameter and it's corresponding value.
                The type of the value can be one of `bool`, `string`, `float`,
                `int`, or `None`.
            metric_dict :
                Each key-value pair in the dictionary is the
                name of the metric and it's corresponding value. Note that the key used
                here should be unique in the tensorboard record. Otherwise the value
                you added by ``add_scalar`` will be displayed in hparam plugin. In most
                cases, this is unwanted.
            hparam_domain_discrete:
                A dictionary that
                contains names of the hyperparameters and all discrete values they can hold
            run_name :
                Name of the run, to be included as part of the logdir.
                If unspecified, will use current timestamp.
            scalars_lists :
                The lists for the loss and accuracy plots.
                This is a list with two lists
                (one for accuracy and one for the loss).
                Each one of the inner lists contain the
                pairs (metric_value, epoch).
            best_not_last:
                boolean flag to decide what value to store in the
                tensorboard tables for the whole training cycle
        Examples::
            from torch.utils.tensorboard import SummaryWriter
            with GiottoSummaryWriter() as w:
                for i in range(5):
                    w.add_hparams({'lr': 0.1*i, 'bsize': i},
                                  {'hparam/accuracy': 10*i, 'hparam/loss': 10*i})
        """
        torch._C._log_api_usage_once("tensorboard.logging.add_hparams")  # type: ignore
        if (not isinstance(hparam_dict, dict)) or (not isinstance(metric_dict, dict)):
            raise TypeError("hparam_dict and metric_dict should be dictionary.")
        exp, ssi, sei = hparams(hparam_dict, metric_dict, hparam_domain_discrete)
        if not run_name:
            run_name = str(time.time()).replace(":", "-")
        logdir = os.path.join(self._get_file_writer().get_logdir(), run_name)
        with SummaryWriter(log_dir=logdir) as w_hp:
            w_hp.file_writer.add_summary(exp)  # type: ignore
            w_hp.file_writer.add_summary(ssi)  # type: ignore
            w_hp.file_writer.add_summary(sei)  # type: ignore
            if isinstance(scalars_lists, list) or isinstance(scalars_lists, tuple):
                scalars_list_loss = scalars_lists[0]
                if best_not_last:
                    scalars_list_loss = [
                        x
                        for x in scalars_list_loss[
                            : int(np.argmin(np.array(scalars_list_loss)[:, 0])) + 1
                        ]
                    ]
                for v, t in scalars_list_loss:
                    w_hp.add_scalar("loss", v, t)
                scalars_list_acc = scalars_lists[1]
                if best_not_last:
                    scalars_list_acc = [
                        x
                        for x in scalars_list_acc[
                            : int(np.argmax(np.array(scalars_list_acc)[:, 0])) + 1
                        ]
                    ]
                for v, t in scalars_list_acc:
                    w_hp.add_scalar("accuracy", v, t)


class HyperParameterOptimization(Trainer):
    """This is the generic class that allows
    the user to perform hyperparameter search over several
    parameters such as learning rate, optimizer.
    Args:
        obj :
            either a Trainer or a Benchmark class
            instance
        search_metric :
            either ``'loss'`` or ``'accuracy'``
        n_trials :
            number of total search trials
        best_not_last :
            boolean flag that is ``True`` would use
            the best metric over epochs averaged over the folds in CV
            rather than the last value of the metrics
            over the epochs averaged over the folds
        pruner :
            Instance of an optuna pruner, can be user-defined
        sampler :
            If left unspecified, ``TPESample`` is used during single-objective
            optimization and ``NSGAIISampler`` during multi-objective optimization
        db_url:
            name of the database to connect to. For example
            ``mysql+mysqldb://usr:psw@host:port/db_name``
        study_name:
            name of the optuna study
    Examples::

        from gdeep.search import HyperParameterOptimization
        # initialise hpo, you need a `trainer`!
        search = HyperParameterOptimization(trainer, "accuracy", 2, best_not_last=True)
        # if you want to store pickle files of the models instead of the state_dicts
        search.store_pickle = True
        # dictionaries of hyperparameters
        optimizers_params = {"lr": [0.001, 0.01]}
        dataloaders_params = {"batch_size": [32, 64, 16]}
        models_hyperparams = {"n_nodes": ["200"]}
        # starting the HPO
        search.start(
            (SGD, Adam),
            3,
            False,
            optimizers_params,
            dataloaders_params,
            models_hyperparams,
            n_accumulated_grads=2,
        )
    """

    is_pipe: bool
    df_res: pd.DataFrame
    study: Study

    def __init__(
        self,
        obj: Union[Trainer, Benchmark],
        search_metric: Literal["loss", "accuracy"] = "loss",
        n_trials: int = 10,
        best_not_last: bool = False,
        pruner: Optional[BasePruner] = None,
        sampler=None,
        db_url: Optional[str] = None,
        study_name: Optional[str] = None,
    ):
        self.best_not_last_gs = best_not_last
        self.best_val_acc_gs = 0.0
        self.best_val_loss_gs = np.inf
        self.list_res: List[Any] = []
        self.db_url = db_url
        self.study_name = study_name
        if isinstance(obj, Trainer):
            self.pipe: Trainer = obj
            super().__init__(
                self.pipe.model,
                self.pipe.dataloaders,
                self.pipe.loss_fn,
                self.pipe.writer,
                self.pipe.training_metric,
                self.pipe.k_fold_class,
            )
            # Pipeline.__init__(self, obj.model, obj.dataloaders, obj.loss_fn, obj.writer)
            self.is_pipe = True
        elif isinstance(obj, Benchmark):
            self.bench: Benchmark = obj
            self.is_pipe = False
        self.search_metric = search_metric
        assert (
            self.search_metric in SEARCH_METRICS
        ), "Wrong search_metric! Either `loss` or `accuracy`"
        self.n_trials = n_trials
        self.val_epoch = 0
        self.train_epoch = 0
        self.sampler = sampler
        if pruner is not None:
            self.pruner = pruner
        else:
            self.pruner = MedianPruner(
                n_startup_trials=5, n_warmup_steps=0, interval_steps=1, n_min_trials=1
            )
        self.scalars_dict: Dict[str, Any] = dict()
        # can be changed by changing this attribute
        self.store_pickle: bool = False

    def _initialise_new_model(
        self, models_hyperparam: Optional[Dict[str, Any]]
    ) -> torch.nn.Module:
        """private method to find the maximal compatible set
        between models and hyperparameters

        Args:
            models_hyperparam (dict):
                model selected hyperparameters

        Returns:
            nn.Module
                torch nn.Module
        """
        new_model = None
        if models_hyperparam:
            list_of_params_keys = HyperParameterOptimization._powerset(
                list(models_hyperparam.keys())
            )
            list_of_params_keys.reverse()
            for params_keys in list_of_params_keys:
                sub_models_hyperparam = {
                    k: models_hyperparam[k]
                    for k in models_hyperparam.keys()
                    if k in params_keys
                }
                try:
                    # print(sub_models_hyperparam)
                    new_model = type(self.model)(**sub_models_hyperparam)  # noqa
                    # print(new_model.state_dict())
                    raise ValueError
                except TypeError:  # when the parameters do not match the model
                    pass
                except ValueError:  # when the parameters match the model
                    break
        else:
            new_model = type(self.model)()

        if new_model is not None:
            warnings.warn("Model cannot be re-initialized. Using existing one.")
            new_model = self.model
        assert (
            new_model is not None
        ), "There is a problem with the re-initialization of the model"
        return new_model

    def _objective(self, trial: BaseTrial, config: HPOConfig):
        """default callback function for optuna's study

        Args:
            trial:
                the independent variable
            config:
                configuration class HPOConfig,
                containing all the parameters needed
        Returns:
            float
                metric (either loss or accuracy)
        """

        # for proper storing of data
        self._cross_validation = config.cross_validation
        self._k_folds = self.k_fold_class.n_splits  # type: ignore
        # generate optimizer
        optimizer = HyperParameterOptimization._new_suggest_categorical(
            trial, "optimizer", config.optimizers
        )

        # generate all the hyperparameters
        self.optimizers_param = HyperParameterOptimization._suggest_params(
            trial, config.optimizers_params
        )
        self.dataloaders_param = HyperParameterOptimization._suggest_params(
            trial, config.dataloaders_params
        )
        self.models_hyperparam = HyperParameterOptimization._suggest_params(
            trial, config.models_hyperparams
        )
        self.schedulers_param = HyperParameterOptimization._suggest_params(
            trial, config.schedulers_params
        )
        # tag for storing the results
        config.writer_tag += "/" + str(
            trial.datetime_start
        )  # str(self.optimizers_param) + \
        # str(self.dataloaders_param) + str(self.models_hyperparam) + \
        # str(self.schedulers_param)
        # create a new model instance
        self.model = self._initialise_new_model(self.models_hyperparam)
        self.pipe = Trainer(
            self.model,
            self.dataloaders,
            self.loss_fn,
            self.writer,
            self.training_metric,
            self.k_fold_class,
        )
        # set best_not_last
        self.pipe.best_not_last = self.best_not_last_gs
        # set the run_name
        self.pipe.run_name = str(trial.datetime_start).replace(":", "-")
        loss, accuracy = self.pipe.train(
            optimizer,
            config.n_epochs,
            config.cross_validation,
            self.optimizers_param,
            self.dataloaders_param,
            config.lr_scheduler,
            self.schedulers_param,
            (trial, self.search_metric),
            config.profiling,
            config.parallel_tpu,
            config.keep_training,
            config.store_grad_layer_hist,
            config.n_accumulated_grads,
            config.writer_tag,
        )

        # dict of the loss per epoch of the training here above. If CV, multiple scalars are stored
        scalars_dict_value = [
            self.pipe.val_loss_list_hparam,
            self.pipe.val_acc_list_hparam,
        ]
        scalar_dict_key = str(trial.datetime_start).replace(":", "-")
        self.scalars_dict[scalar_dict_key] = scalars_dict_value
        # release the run_name
        self.pipe.run_name = None
        self.writer.flush()  # type: ignore
        # print
        self._print_output()

        # save model and optimizer
        save_model_and_optimizer(
            self.pipe.model,
            trial_id=str(trial.datetime_start).replace(":", "-"),
            optimizer=self.pipe.optimizer,
            store_pickle=self.store_pickle,
        )

        # save results to tensorboard
        model_name, dataset_name = self._extract_model_and_dataset_name(
            config.writer_tag
        )

        self._store_trial_to_tb(
            trial,
            scalars_dict_value,
            scalar_dict_key,
            model_name,
            dataset_name,
            loss,
            accuracy,
        )

        # returns
        if self.search_metric == "loss":
            self.best_val_acc_gs = max(self.best_val_acc_gs, accuracy)
            self.best_val_loss_gs = min(self.best_val_loss_gs, loss)
            return loss
        else:
            self.best_val_acc_gs = max(self.best_val_acc_gs, accuracy)
            self.best_val_loss_gs = min(self.best_val_loss_gs, loss)
            return accuracy

    def _extract_model_and_dataset_name(self, writer_tag: str) -> Tuple[str, str]:
        """Extract the model and dataset from the writer_tag"""
        index_ds = writer_tag.find("Dataset:")
        if index_ds == -1:
            dataset_name = self.pipe.dataloaders[0].dataset.__class__.__name__
        else:
            dataset_name = writer_tag[index_ds + 8 : writer_tag.find("|Model:")]
        index_md = writer_tag.find("|Model:")
        if index_md == -1:
            model_name = self.pipe.model.__class__.__name__
        else:
            model_name = writer_tag[index_md + 7 : writer_tag.find("/")]
        return model_name, dataset_name

    def start(
        self,
        optimizers: List[Type[Optimizer]],
        n_epochs: int = 1,
        cross_validation: bool = False,
        optimizers_params: Optional[Dict[str, Any]] = None,
        dataloaders_params: Optional[Dict[str, Any]] = None,
        models_hyperparams: Optional[Dict[str, Any]] = None,
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        schedulers_params: Optional[Dict[str, Any]] = None,
        profiling: bool = False,
        parallel_tpu: bool = False,
        keep_training: bool = False,
        store_grad_layer_hist: bool = False,
        n_accumulated_grads: int = 0,
        writer_tag: str = "",
    ) -> None:
        """method to be called when starting the hyperparameter optimization
        Args:
            optimizers:
                list of torch optimizers classes, not instances
            n_epochs:
                number of training epochs
            cross_validation:
                whether or not to use cross-validation
            optimizers_params:
                dictionary of optimizers params
            dataloaders_params:
                dictionary of dataloaders parameters
            models_hyperparams:
                dictionary of model parameters
            lr_scheduler:
                torch learning rate scheduler class
            schedulers_params:
                learning rate scheduler parameters
            profiling :
                whether or not you want to activate the
                profiler
            parallel_tpu:
                boolean value to run the computations
                on multiple TPUs
            n_accumulated_grads (int, default=0):
                number of accumulated gradients. It is
                considered only if a positive integer
            keep_training:
                bool flag to decide whether to continue
                training or not
            store_grad_layer_hist:
                flag to store the gradients of the layers in the
                tensorboard histograms
            writer_tag:
                tag to prepend to the output
                on tensorboard
        """
        if self.search_metric == "loss":
            self.study = optuna.create_study(
                direction="minimize",
                sampler=self.sampler,
                pruner=self.pruner,
                storage=self.db_url,
                study_name=self.study_name
                if self.study_name is not None
                else "".join(
                    random.choice(string.ascii_uppercase + string.digits)
                    for _ in range(20)
                ),
                load_if_exists=True,
            )
        else:
            self.study = optuna.create_study(
                direction="maximize",
                sampler=self.sampler,
                pruner=self.pruner,
                storage=self.db_url,
                study_name=self.study_name
                if self.study_name is not None
                else "".join(
                    random.choice(string.ascii_uppercase + string.digits)
                    for _ in range(20)
                ),
                load_if_exists=True,
            )
        config = HPOConfig(
            optimizers,
            n_epochs,
            cross_validation,
            optimizers_params,
            dataloaders_params,
            models_hyperparams,
            lr_scheduler,
            schedulers_params,
            profiling,
            parallel_tpu,
            keep_training,
            store_grad_layer_hist,
            n_accumulated_grads,
            writer_tag,
        )

        if self.is_pipe:
            # in the __init__, self.model and self.dataloaders are
            # already initialized. So they exist also in _objective()
            self._inner_optimization_fun(self.model, self.dataloaders, config)

        else:
            _benchmarking_param(
                self._inner_optimization_fun,
                (self.bench.models_dicts, self.bench.dataloaders_dicts),
                config,
            )
        # self._store_to_tensorboard()

    def _inner_optimization_fun(
        self,
        model: Union[torch.nn.Module, Dict[str, torch.nn.Module]],
        dataloaders: Union[
            List[DataLoader[Tuple[Union[Tensor, List[Tensor]], Tensor]]],
            Dict[str, DataLoader],
        ],
        config: HPOConfig,
    ) -> None:
        """private method to be decorated with the
        benchmark decorator to have benchmarking
        or simply used as is if no benchmarking is
        needed
        """

        try:
            config.writer_tag = (
                "Dataset:" + dataloaders["name"] + "|Model:" + model["name"]  # type: ignore
            )
            super().__init__(
                model["model"],  # type: ignore
                dataloaders["dataloaders"],  # type: ignore
                self.bench.loss_fn,
                self.bench.writer,
                self.bench.training_metric,
                self.bench.k_fold_class,
            )
        except TypeError:
            pass

        self.study.optimize(  # type: ignore
            lambda tr: self._objective(tr, config), n_trials=self.n_trials, timeout=None
        )

        try:
            self._results(model_name=model["name"], dataset_name=dataloaders["name"])  # type: ignore
            # save_model_and_optimizer(self.pipe.model,
            #                         model["name"] +
            #                         str(self.optimizers_param) +
            #                         str(self.dataloaders_param) +
            #                         str(self.models_hyperparam) +
            #                         str(self.schedulers_param),
            #                         self.pipe.optimizer)
        except TypeError:
            try:
                self._results(
                    model_name=self.pipe.model.__class__.__name__,
                    dataset_name=self.pipe.dataloaders[0].dataset.__class__.__name__,
                )
                # save_model_and_optimizer(self.pipe.model,
                #                         optimizer=self.pipe.optimizer)
            except AttributeError:
                self._results()

    def _print_output(self) -> None:
        """Printing the results of an optimization"""
        results_string_to_print = (
            ("\nBest Validation loss: " + str(self.pipe.best_val_loss))
            if self.search_metric == "loss"
            else ("\nBest Validation accuracy: " + str(self.pipe.best_val_acc))
        )

        string_to_print = (
            "\nModel Hyperparameters: "
            + str(self.models_hyperparam)
            + "\nOptimizer: "
            + str(self.pipe.optimizer)
            + "\nOptimizer parameters: "
            + str(self.optimizers_param)
            + "\nDataloader parameters: "
            + str(self.dataloaders_param)
            + "\nLR-scheduler parameters: "
            + str(self.schedulers_param)
            + results_string_to_print
        )
        try:
            # print models, metric and hyperparameters
            print(
                "*" * 20 + " RESULTS " + 20 * "*" + "\n",
                "\nModel: ",
                self.pipe.model.__class__.__name__,
                string_to_print,
            )
        except AttributeError:
            print("*" * 20 + " RESULTS " + 20 * "*" + "\n" + string_to_print)

    def _store_to_list_each_step(
        self,
        trial: BaseTrial,
        model_name: str,
        dataset_name: str,
        loss: Optional[float] = np.inf,
        accuracy: Optional[float] = -1.0,
        list_res: Optional[List[Any]] = None,
    ) -> List[Any]:
        """Private method to store all the HPO parameters
        of one trial

        Args:
            trial (optuna.trial):
                the trial at hand
            model_name (str):
                name of the model
            dataset_name (str):
                name of the dataset
            loss (float, default np.inf)
                the value of the loss for the current trial
            accuracy (float, default -1):
                the value of the accuracy for te current trial
            list_res (list):
                list of results
        Returns:
            list:
                list of HPOs and metrics. the first element is the
                run name and the last two are the metrics (loss
                and accuracy)
        """
        if list_res is None:
            list_res = []
        temp_list = []
        for val in trial.params.values():
            temp_list.append(val)
        if self.search_metric == "loss":
            try:
                loss = trial.value  # type: ignore
            except AttributeError:
                loss = loss
            list_res.append(
                [str(trial.datetime_start).replace(":", "-"), model_name, dataset_name]
                + temp_list
                + [loss, -1]  # type: ignore
            )
        else:
            try:
                accuracy = trial.value  # type: ignore
            except AttributeError:
                accuracy = accuracy
            list_res.append(
                [str(trial.datetime_start).replace(":", "-"), model_name, dataset_name]
                + temp_list
                + [np.inf, accuracy]  # type: ignore
            )
        return list_res

    def _store_trial_to_tb(
        self,
        trial: BaseTrial,
        scalars_dict_value: List[List[List[float]]],
        scalar_dict_key: str,
        model_name: str,
        dataset_name: str,
        loss: float,
        accuracy: float,
    ) -> None:
        """store hyperparameters of a single trial to
        tensorboard
        """
        list_res = self._store_to_list_each_step(
            trial, model_name, dataset_name, loss, accuracy
        )

        keys = ["model", "dataset"] + list(trial.params.keys())

        # average over the scalars_dict
        scalars_dict_avg = self._refactor_scalars(scalars_dict_value)
        # dict of parameters
        dictionary_ = {
            k: (int(v) if isinstance(v, np.int64) or isinstance(v, np.int32) else v)  # type: ignore
            for k, v in zip(keys, list_res[0][1:-2])
        }

        try:
            self.writer.add_hparams(  # type: ignore
                dictionary_,
                {"loss": list_res[0][-2], "accuracy": list_res[0][-1]},
                run_name=scalar_dict_key,
                scalars_lists=scalars_dict_avg,  # type: ignore
                best_not_last=self.best_not_last_gs,  # type: ignore
            )
        except KeyError:  # this happens when trials have been pruned
            pass

        self.writer.flush()  # type: ignore

    def _results(
        self, model_name: str = "model", dataset_name: str = "dataset"
    ) -> pd.DataFrame:
        """This method returns the dataframe with all the results of
        the hyperparameters optimization.
        It also saves the figures in the writer.
        Args:
            model_name:
                the model name for the
                tensorboard hpo table
            dataset_name:
                the dataset name for the
                tensorboard hpo table
        Returns:
            pd.DataFrame:
                the hyperparameter table
        """
        self.list_res = []
        trials = self.study.trials
        pruned_trials = self.study.get_trials(
            deepcopy=False, states=(TrialState.PRUNED,)
        )
        complete_trials = self.study.get_trials(
            deepcopy=False, states=(TrialState.COMPLETE,)
        )

        print("Study statistics: ")
        print("Number of pruned trials: ", len(pruned_trials))
        print("Number of complete trials: ", len(complete_trials))
        try:
            print("******************** BEST TRIAL: ********************")
            trial_best = self.study.best_trial
            print("Metric Value for best trial: ", trial_best.value)
            print("Parameters Values for best trial: ", trial_best.params)
            print("DateTime start of the best trial: ", trial_best.datetime_start)
        except ValueError:
            warnings.warn("No best trial found. Using the first trial")
            trial_best = self.study.trials[0]

        for trial in trials:
            self._store_to_list_each_step(
                trial, model_name, dataset_name, trial.value, trial.value, self.list_res
            )

        self.df_res = pd.DataFrame(
            self.list_res,
            columns=["run_name", "model", "dataset"]
            + list(trial_best.params.keys())
            + ["loss", "accuracy"],
        )
        # compute hyperparams correlation
        corr, labels = self._correlation_of_hyperparams()
        if self.n_trials > 1:
            try:
                fig2 = px.imshow(
                    corr,
                    labels=dict(x="Parameters", y="Parameters", color="Correlation"),
                    x=labels,
                    y=labels,
                )
                fig2.update_xaxes(side="top")
                # fig2.show()
                img2 = plotly2tensor(fig2)

                self.writer.add_images(  # type: ignore
                    "HPO correlation: " + model_name + " " + dataset_name,
                    img2,
                    dataformats="HWC",
                )
                self.writer.flush()  # type: ignore
            except ValueError:
                warnings.warn(
                    "Cannot send the picture of the correlation"
                    + " matrix to tensorboard"
                )

        return self.df_res

    def _correlation_of_hyperparams(self) -> Tuple[Array, List[Any]]:
        """Correlations of numerical hyperparameters"""
        list_of_arrays = []
        labels = []
        for col in self.df_res.columns:
            vals = self.df_res[col].values
            # print("type here", vals.dtype)
            if vals.dtype in [
                np.float16,
                np.float64,
                np.float32,
                np.int32,
                np.int64,
                np.int16,
            ]:
                list_of_arrays.append(vals)
                labels.append(col)
        # print(list_of_arrays)
        with KnownWarningSilencer():
            corr = np.corrcoef(np.array(list_of_arrays))
        return corr, labels

    def _refactor_scalars(
        self, two_lists: List[List[List[float]]]
    ) -> Tuple[List[Any], List[Any]]:
        """private method to transform a list with
        many values for the same epoch into a dictionary
        compatible with ``add_scalar`` averaged per epoch
        Args:
            two_lists :
                two lists with pairs (value, time) with possible
                repetition of the same time
        Returns:
            list of list:
                compatible with ``add_scalar``
        """

        out0 = _inner_refactor_scalars(
            two_lists[0], self._cross_validation, self._k_folds
        )
        out1 = _inner_refactor_scalars(
            two_lists[1], self._cross_validation, self._k_folds
        )
        return out0, out1

    @staticmethod
    def _suggest_params(
        trial: BaseTrial, params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """Utility function to generate the parameters
        for the hyperparameter search. It is based on optuna `suggest_<type>`.
        Args:
            trial (optuna.trial):
                optuna trial variable
            params (dict):
                dictionary of parameters
        Returns:
            (dict):
                dictionary of selected parameters values
        """
        if params is None:
            return None
        for k, v in params.items():
            if (isinstance(v, list) or isinstance(v, tuple)) and len(v) == 1:
                params[k] = 2 * v
        # param_temp = {}
        param_temp = {
            k: HyperParameterOptimization._new_suggest_float(trial, k, *v)
            for k, v in params.items()
            if (isinstance(v, list) or isinstance(v, tuple))
            and (isinstance(v[0], float) or isinstance(v[1], float))
        }
        param_temp2 = {
            k: trial.suggest_int(k, *v)
            for k, v in params.items()
            if (isinstance(v, list) or isinstance(v, tuple))
            and (isinstance(v[0], int) or isinstance(v[1], int))
        }
        param_temp.update(param_temp2)
        param = {
            k: HyperParameterOptimization._new_suggest_categorical(trial, k, v)
            for k, v in params.items()
            if (isinstance(v, list) or isinstance(v, tuple))
            and not (isinstance(v[0], int) or isinstance(v[1], int))
            and not (isinstance(v[0], float) or isinstance(v[1], float))
        }
        param.update(param_temp)
        # print("suggested parameters:", param)
        return param

    @staticmethod
    def _new_suggest_categorical(trial, name: str, choices: Sequence) -> Any:
        """A modification of the Optuna function, in order to remove the
        constraints on the type for categorical choices.
        Expected to get a list with items of the same type (or base class).
        For example, ``choices`` being a list of Callables or classes would
        work.
        """

        if isinstance(choices, list) or isinstance(choices, tuple):
            if (
                isinstance(choices[0], str)
                or isinstance(choices[0], int)
                or isinstance(choices[0], float)
            ):
                return trial.suggest_categorical(name, choices)
            else:  # in case optuna cannot handle the types
                dict_choices = {x.__name__: x for x in choices}
                key = trial.suggest_categorical(
                    name, dict_choices.keys()
                )  # random choice on the names
                return dict_choices[key]  # return the corresponding value
        return random.choice(choices)

    @staticmethod
    def _new_suggest_float(trial, name, low, high, step=None, log=False):
        """A modification of the Optuna function, in order to remove the `*`"""
        return trial.suggest_float(name, low, high, log=log, step=step)

    @staticmethod
    def _powerset(iterable):
        """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
        s = list(iterable)
        return list(chain.from_iterable(combinations(s, r) for r in range(len(s) + 1)))



from typing import Optional

import torch
from torch.nn import Dropout, MultiheadAttention

from ..persformer_config import PersformerConfig
from .attention_base import AttentionBase

# Type aliases
from gdeep.utility.custom_types import Tensor


class ScaledDotProductAttention(AttentionBase):
    """
    Dot product attention. See https://arxiv.org/abs/1706.03762.
    """

    def __init__(self, config: PersformerConfig):
        super().__init__(config)

        self.scaled_dot_product_attention = MultiheadAttention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_probs_dropout_prob,
            batch_first=True,
        )
        self.dropout = Dropout(config.hidden_dropout_prob)

    def forward(
        self, input: Tensor, attention_mask: Optional[Tensor] = None  # type: ignore
    ):
        """
        Forward pass.
        """
        # if attention_mask is not None:
        #     attention_mask = torch.stack([attention_mask] * attention_mask.shape[-1], dim=-1)
        #     attn_mask = torch.concat([attention_mask] * self.config.num_attention_heads, dim=0)
        # else:
        #     attn_mask = None
        attention_output, _ = self.scaled_dot_product_attention(
            input, input, input, key_padding_mask=attention_mask
        )
        return self.dropout(attention_output)



from enum import Enum, auto


class PoolerType(Enum):
    ATTENTION = auto()
    MAX = auto()
    MEAN = auto()
    SUM = auto()


class LayerNormStyle(Enum):
    """
    The style of layer normalization.
    """

    NO_LAYER_NORMALIZATION = auto()
    PRE_LAYER_NORMALIZATION = auto()
    POST_LAYER_NORMALIZATION = auto()


class AttentionType(Enum):
    """
    The type of attention.
    """

    NO_ATTENTION = auto()
    DOT_PRODUCT = auto()
    UNNORMALIZED_DOT_PRODUCT = auto()
    INDUCED_ATTENTION = auto()
    FOURIER_MIXER = auto()


class ActivationFunction(Enum):
    """
    The activation function.
    """

    RELU = auto()
    GELU = auto()
    SELU = auto()
    MISH = auto()



from typing import Any, List, Tuple, Union

import torch
from torch.utils.data import Dataset
from torchvision.transforms import Resize, ToTensor

from ..abstract_preprocessing import AbstractPreprocessing

# type definition
from gdeep.utility.custom_types import Tensor


class ToTensorImage(AbstractPreprocessing[Tuple[Any, Any], Tuple[Tensor, Tensor]]):
    """Class to preprocess image files for classification
    tasks

        Args:
            size :
                Desired output size. If size is a sequence like (h, w),
                output size will be matched to this. If size is an int,
                smaller edge of the image will be matched to this number.
                I.e, if height > width, then image will be rescaled to
                ``(size * height / width, size)``.

    """

    is_fitted: bool
    size: Union[int, List[int]]

    def __init__(self, size: Union[int, List[int]]) -> None:
        self.size = size
        self.is_fitted = True

    def fit_to_dataset(self, dataset: Dataset[Any]) -> None:
        # nothing to do
        pass

    def __call__(self, datum: Tuple[Any, Any]) -> Tuple[Tensor, Tensor]:
        return ToTensor()(Resize(self.size)(datum[0])), torch.tensor(datum[1], dtype=torch.long)  # type: ignore



from typing import Callable, Dict

from gdeep.utility.enum_types import AttentionType

from .attention import AttentionBase, InducedAttention, ScaledDotProductAttention
from .persformer_config import PersformerConfig


class AttentionFactory:
    """
    Factory for creating attention modules.
    """

    attention_modules: Dict[
        AttentionType, Callable[[PersformerConfig], AttentionBase]
    ] = {}

    def __init__(self):
        # Register the attention layers here:
        self.register_attention__builder(
            AttentionType.DOT_PRODUCT, lambda config: ScaledDotProductAttention(config)
        )
        self.register_attention__builder(
            AttentionType.INDUCED_ATTENTION, lambda config: InducedAttention(config)
        )

    def register_attention__builder(
        self,
        attention_type: AttentionType,
        attention_module_builder: Callable[[PersformerConfig], AttentionBase],
    ) -> None:
        """
        Register an attention module.
        """
        self.attention_modules[attention_type] = attention_module_builder

    def build(self, config: PersformerConfig) -> AttentionBase:
        """
        Create an attention module.
        """
        return self.attention_modules[config.attention_type](config)



from typing import List, Tuple

import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from torch.optim import SGD
import numpy as np

from gdeep.trainer import Trainer
from gdeep.models import FFNet
from gdeep.search import GiottoSummaryWriter
from gdeep.data.datasets import FromArray, DataLoaderBuilder
from gdeep.search import clean_up_files

from gdeep.utility.custom_types import Tensor


class MyDataset(Dataset):
    def __init__(self):
        self.x = []
        for _ in range(100):
            self.x.append((torch.rand(1, np.random.randint(2, 4))).to(torch.float))
        self.y = np.array(
            np.random.randint(2, size=100 * 2).reshape(-1, 2), dtype=np.int64
        )

    def __len__(self):
        return 100

    def __getitem__(self, item: int) -> Tuple[Tensor, Tensor]:
        return self.x[item], self.y[item]


def collate_fn(batch_tuple: List):
    # print(batch_tuple)
    target = torch.zeros(len(batch_tuple), 4)
    label = torch.zeros(len(batch_tuple), 2).to(torch.long)
    for i, batch in enumerate(batch_tuple):
        source = batch[0]
        target[i, : len(source[-1])] = source
        label[i] = torch.tensor(batch[1]).to(torch.long)
    return target, label


# model
class Model1(nn.Module):
    def __init__(self):
        super(Model1, self).__init__()
        self.seqmodel = nn.Sequential(nn.Flatten(), FFNet(arch=(4, 5, 4)))

    def forward(self, x):
        return self.seqmodel(x).reshape(-1, 2, 2)


@clean_up_files
def test_trainer_from_array():

    model = Model1()
    # dataloaders
    X = np.array(np.random.rand(100, 4), dtype=np.float32)  # type: ignore
    y = np.array(np.random.randint(2, size=100 * 2).reshape(-1, 2), dtype=np.int64)
    dl_tr, *_ = DataLoaderBuilder([FromArray(X, y)]).build([{"batch_size": 23}])

    # loss function
    loss_fn = nn.CrossEntropyLoss()
    # tb writer
    writer = GiottoSummaryWriter()
    # pipeline
    pipe = Trainer(model, [dl_tr, None], loss_fn, writer)  # type: ignore
    # then one needs to train the model using the pipeline!
    pipe.train(SGD, 2, True, {"lr": 0.001}, n_accumulated_grads=2)


def test_trainer_collate():

    model = Model1()
    # dataloaders
    ds = MyDataset()
    dl_tr = DataLoader(ds, batch_size=6, collate_fn=collate_fn)

    # loss function
    loss_fn = nn.CrossEntropyLoss()
    # tb writer
    writer = GiottoSummaryWriter()
    # pipeline
    pipe = Trainer(model, [dl_tr, None], loss_fn, writer)  # type: ignore
    # then one needs to train the model using the pipeline!
    pipe.train(SGD, 2, True, {"lr": 0.001}, n_accumulated_grads=2)

    # evaluation
    assert len(pipe.evaluate_classification(2)) == 3



from gdeep.search import HPOConfig
from torch.optim import SGD


def test_hpo_config() -> None:
    config = HPOConfig([SGD])
    dictionary = config.to_dict()
    assert isinstance(dictionary, dict)



from typing import Dict, Any


# Define the dataset factory class for the tori dataset and torchvision datasets
# using the factory design pattern
# https://realpython.com/factory-method-python/


class DatasetFactory(object):
    """Dataset factory class for the tori dataset and torchvision datasets
    using the factory design pattern

    Examples::

        # Create a dataset for the tori dataset
        dataset = get_dataset("Tori", name="DoubleTori", n_points=100)

        # Create the MNIST dataset
        dataset = get_dataset("Torchvision", name="MNIST")

    """

    _builders: Dict[str, Any]

    def __init__(self):
        self._builders = {}

    def register_builder(self, key: str, builder: Any):
        """this method adds to the internal builders dictionary
        new dataloader builders
        """
        self._builders[key] = builder  # type: ignore

    def build(self, key: str, **kwargs) -> Any:  # type: ignore
        """This method returns the DataLoader builder
        corresponding to the input key.

        Args:
            key:
                the name of the dataset
        """
        builder = self._builders.get(key)  # type: ignore
        if builder is None:
            raise ValueError(f"No builder registered for key {key}")
        return builder(**kwargs)  # type: ignore



import os
import time

from PIL import Image
import numpy as np
from plotly.io import write_image
import torch


def plotly2tensor(fig) -> torch.Tensor:
    """convert Plotly fig to an array.

    Args:
        fig (plotly GraphicObject):
            figure to convert to tensor

    Returns:
        Tensor:
            the tensor discretisation of the
            figure
    """
    now = str(time.time()).replace(".", "-")
    try:
        write_image(fig, "deleteme" + now + ".jpeg", format="jpeg", engine="orca")
    except ValueError:
        write_image(fig, "deleteme" + now + ".jpeg", format="jpeg")
    with Image.open("deleteme" + now + ".jpeg") as img:
        arr = np.asarray(img).copy()  # type: ignore
    os.remove("deleteme" + now + ".jpeg")
    return torch.from_numpy(arr)


def png2tensor(file_name) -> torch.Tensor:
    img = Image.open(file_name)
    arr = np.asarray(img).copy()  # type: ignore
    os.remove(file_name)
    return torch.from_numpy(arr)



import warnings
from typing import TypeVar, Tuple, Any

# Check if python version is 3.10 or higher
from matplotlib.pyplot import axis

import torch
from torch.utils.data import Dataset

from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from gdeep.utility.custom_types import Tensor

from ..abstract_preprocessing import AbstractPreprocessing
from ..transforming_dataset import TransformingDataset


PD = OneHotEncodedPersistenceDiagram

T = TypeVar("T")


class Normalization(AbstractPreprocessing[Tuple[Tensor, T], Tuple[Tensor, T]]):
    """This class runs the standard normalisation on all the dimensions of
    the tensors of a dataset. For example, in case of images where each item is of
    shape ``(C, H, W)``, the average will and the standard deviations
    will be tensors of shape ``(C, H, W)``
    """

    is_fitted: bool
    mean: Tensor
    stddev: Tensor

    def __init__(self):
        self.is_fitted = False

    def fit_to_dataset(self, dataset: Dataset[Tuple[Tensor, T]]) -> None:
        self.mean = _compute_mean_of_dataset(dataset)
        self.stddev = _compute_stddev_of_dataset(dataset, self.mean)
        self.is_fitted = True

    def __call__(self, item: Tuple[Tensor, T]) -> Tuple[Tensor, T]:
        if not self.is_fitted:
            raise RuntimeError(
                "The normalisation is not fitted to any dataset."
                " Please call fit_to_dataset() first."
            )
        if not torch.all(self.stddev > 0):
            warnings.warn("The standard deviation contains zeros! Adding 1e-7")
            self.stddev = self.stddev + 1e-7
        out = (item[0] - self.mean) / self.stddev
        return (out, item[1])


def _compute_mean_of_dataset(dataset: Dataset[Tuple[Tensor, Any]]) -> Tensor:
    """Compute the mean of the whole dataset"""
    mean: Tensor = torch.zeros(dataset[0][0].shape, dtype=torch.float64)
    for idx in range(len(dataset)):  # type: ignore
        if idx == 0:
            mean += dataset[idx][0]
        else:
            mean = (mean * idx + dataset[idx][0]) / (idx + 1)
    return mean


def _compute_stddev_of_dataset(
    dataset: Dataset[Tuple[Tensor, Any]], mean: Tensor
) -> Tensor:
    """Compute the stddev of the whole dataset"""

    def square_diff_from_mean(x: Tuple[Tensor, Any]) -> Tuple[Tensor, Any]:
        return (x[0] - mean) ** 2, x[1]

    mean_normalized_dataset = TransformingDataset[
        Tuple[Tensor, Any], Tuple[Tensor, Any]
    ](dataset, square_diff_from_mean)
    stddev: Tensor = _compute_mean_of_dataset(mean_normalized_dataset)
    return stddev.sqrt()



from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar

import numpy as np
import plotly.graph_objs as gobj
import torch
from gdeep.utility.utils import flatten_list_of_lists

T = TypeVar("T")
from gdeep.utility.custom_types import Tensor

from gdeep.utility.custom_types import Array


class OneHotEncodedPersistenceDiagram:
    """This class represents a single one-hot encoded persistence diagram.

    Args:
        data:
            The data of the persistence diagram. The data must be a tensor of shape
            (num_points, 2 + num_homology_dimensions) and the last dimension must be
            the concatenation of the birth-death-coordinates and the one-hot encoded homology
            dimension.
            The invariants of the persistence diagram are checked in the constructor.
        homology_dimension_names:
            The names of the homology dimensions. If None, the names are set to H_0, H_1, ...

    Example::
            pd =  torch.tensor\
                      ([[0.0928, 0.0995, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0916, 0.1025, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0978, 0.1147, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0978, 0.1147, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0916, 0.1162, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0740, 0.0995, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0728, 0.0995, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0740, 0.1162, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0728, 0.1162, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0719, 0.1343, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0830, 0.2194, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0830, 0.2194, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0719, 0.2194, 0.0000, 1.0000, 0.0000, 0.0000]])

            names = ["Ord0", "Ext0", "Rel1", "Ext1"]
            pd = OneHotEncodedPersistenceDiagram(pd, names)
    """

    _data: Tensor
    _homology_dimension_names: List[str]

    def __init__(self, data: Tensor, homology_dimension_names: Optional[List[str]] = None):  # type: ignore
        _check_if_valid(data)
        self._data = data
        if homology_dimension_names is not None:
            self._homology_dimension_names = homology_dimension_names
        else:
            self._homology_dimension_names = [
                "H_" + str(i) for i in range(self.get_num_homology_dimensions())
            ]
        assert (
            len(self._homology_dimension_names) == self.get_num_homology_dimensions()
        ), "The number of names must be equal to the number of homology dimensions."

    def set_homology_dimension_names(self, homology_dimension_names: List[str]) -> None:
        """This method sets the homology dimension names."""
        assert (
            len(homology_dimension_names) == self.get_num_homology_dimensions()
        ), "The number of names must be equal to the number of homology dimensions."
        self._homology_dimension_names = homology_dimension_names

    def get_num_homology_dimensions(self) -> int:
        """This method returns the number of homology dimensions."""
        return self._data.shape[-1] - 2

    def get_num_points(self) -> int:
        """This method returns the number of points."""
        return self._data.shape[0]

    def __repr__(self):
        return (
            f"OneHotEncodedPersistenceDiagram({self._data.shape})\n"
            f"{self.get_num_homology_dimensions()} homology dimensions\n"
            f"{self._data.shape[0]} points\ndata:\n {self._data.__repr__()}"
        )

    def save(self, path: str) -> None:
        """This method saves the persistence diagram to a file."""
        torch.save(self._data, path)

    def get_points_in_homology_dimension(
        self, homology_dimension: int
    ) -> "OneHotEncodedPersistenceDiagram":
        """This method returns all points in a given homology dimension."""
        assert (
            0 <= homology_dimension < self.get_num_homology_dimensions()
        ), "The homology dimension must be smaller than the number of homology dimensions."
        return OneHotEncodedPersistenceDiagram(
            self._data[torch.where(self._data[:, 2 + homology_dimension] == 1.0)]
        )

    def get_raw_data(self) -> Tensor:
        """This method returns the raw data of the persistence diagram.
        This function should not be used to change the data.
        """
        return self._data

    @staticmethod
    def load(path: str) -> "OneHotEncodedPersistenceDiagram":
        """This method loads a persistence diagram from a file."""
        return OneHotEncodedPersistenceDiagram(torch.load(path))

    def get_all_points_in_homology_dimension(
        self, homology_dimension: int
    ) -> "OneHotEncodedPersistenceDiagram":
        """This method returns all points in a given homology dimension."""
        assert (
            homology_dimension < self.get_num_homology_dimensions()
        ), "The homology dimension must be smaller than the number of homology dimensions."
        return OneHotEncodedPersistenceDiagram(
            self._data[torch.where(self._data[:, 2 + homology_dimension] == 1.0)]
        )

    def plot(self, names: Optional[List[str]] = None) -> gobj.Figure:  # type: ignore
        """This method plots the persistence diagram.

        Args:
            names:
                The names of the homology dimensions.

        Examples::

            pd =  torch.tensor\
                      ([[0.0928, 0.0995, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0916, 0.1025, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0978, 0.1147, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0978, 0.1147, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0916, 0.1162, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0740, 0.0995, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0728, 0.0995, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0740, 0.1162, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0728, 0.1162, 0.0000, 0.0000, 1.0000, 0.0000],
                        [0.0719, 0.1343, 0.0000, 0.0000, 0.0000, 1.0000],
                        [0.0830, 0.2194, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0830, 0.2194, 1.0000, 0.0000, 0.0000, 0.0000],
                        [0.0719, 0.2194, 0.0000, 1.0000, 0.0000, 0.0000]])

            names = ["Ord0", "Ext0", "Rel1", "Ext1"]
            pd = OneHotEncodedPersistenceDiagram(pd, names)
            pd.plot()

        """
        names = self._homology_dimension_names if names is None else names
        assert (
            len(names) == self.get_num_homology_dimensions()
        ), "The number of names must be equal to the number of homology dimensions."
        pd = self._data.detach().numpy()

        # convert one-hot encoding to categorical encoding
        pd_categorical = pd[:, 2:].argmax(axis=1)

        fig = _plot_diagram(
            np.concatenate([pd[:, :2], pd_categorical.reshape(-1, 1)], axis=1),
            names=names,
        )

        return fig

    def all_close(
        self, other: "OneHotEncodedPersistenceDiagram", atol: float = 1e-7
    ) -> bool:
        """This method checks if the persistence diagrams are close."""
        for i in range(self.get_num_homology_dimensions()):
            if not torch.allclose(
                self.get_all_points_in_homology_dimension(i).get_raw_data(),
                other.get_all_points_in_homology_dimension(i).get_raw_data(),
                atol=atol,
            ):
                return False
        return True

    def get_lifetimes(self) -> Tensor:
        """This method returns the lifetimes of the points."""
        return self._data[:, 1] - self._data[:, 0]

    def filter_by_lifetime(
        self, min_lifetime: float, max_lifetime: float
    ) -> "OneHotEncodedPersistenceDiagram":
        """This method filters the persistence diagram by lifetime.

        Args:
            min_lifetime:
                The minimum lifetime of the remaining points.
            max_lifetime:
                The maximum lifetime of the remaining points.
        """
        lifetime: Tensor = self.get_lifetimes()
        return OneHotEncodedPersistenceDiagram(
            self._data[  # type: ignore
                torch.where((lifetime >= min_lifetime) & (lifetime <= max_lifetime))
            ]
        )

    @staticmethod
    def from_numpy(data: Array) -> "OneHotEncodedPersistenceDiagram":
        """This method creates a persistence diagram from a numpy array."""
        # assert data.dtype == np.float32, "The data must be of type np.float32. Otherwise,"\
        #     " the data will not be correctly converted."
        return OneHotEncodedPersistenceDiagram(
            torch.from_numpy(data.astype(np.float32))
        )


def collate_fn_persistence_diagrams(
    batch: List[Tuple[OneHotEncodedPersistenceDiagram, int]]
) -> Tuple[List[Tensor], Tensor]:
    """This function collates the data for the persistence diagram by padding the data, converting
    the data to tensors, converting the labels to tensors and generating masks for the valid
    entries.

    The input is a list of tuples of the form (persistence diagram, label).

    Args:
        batch:
            The list of tuples of the form (persistence diagram, label).

    Returns:
        The data, the labels and the masks.
    """
    max_num_points: int = max([len(x[0].get_raw_data()) for x in batch])
    num_homology_dimensions: int = batch[0][0].get_num_homology_dimensions()
    input_batch = torch.zeros(len(batch), max_num_points, num_homology_dimensions + 2)
    mask = torch.zeros(len(batch), max_num_points)
    for i, (pd, _) in enumerate(batch):
        input_batch[i, : len(pd.get_raw_data()), :] = pd.get_raw_data()
        mask[i, : len(pd.get_raw_data())] = 1.0

    label_batch = torch.tensor([x[1] for x in batch])

    return [input_batch, mask], label_batch


def _check_if_valid(data) -> None:
    """This method checks if the data is a valid persistence diagram.

    Args:
        data:
            The data to check.

    Raises:
        ValueError:
            If the data is not a valid persistence diagram.
    """

    if data.ndimension() != 2:
        raise ValueError(
            "The input should be a 2-dimensional tensor."
            f"The input has {data.ndimension()} dimensions."
        )
    assert data.shape[-1] > 2, "The input should have at least one homology dimensions."
    assert torch.all(data[:, 2:] >= -1e-5) and torch.allclose(
        data[:, 2:].sum(dim=1), torch.tensor(1.0)
    ), "The homology dimension should be one-hot encoded."


def get_one_hot_encoded_persistence_diagram_from_gtda(
    persistence_diagram: Array,
) -> OneHotEncodedPersistenceDiagram:
    """This function takes a single persistence diagram from giotto-tda and returns a one-hot encoded
    persistence diagram.

    Args:
        persistence_diagram:
            An array of shape (num_points, 3) where the first two columns
            represent the coordinates of the points and the third column
            represents the index of the homology dimension.

    Returns:
        OneHotEncodedPersistenceDiagram:
            A one-hot encoded persistence diagram. If the persistence diagram has only one homology
            dimension, the third column will be filled with ones.
    """
    assert (
        persistence_diagram.ndim == 2 and persistence_diagram.shape[1] >= 2
    ), "The input should be a 2-dimensional array of shape (num_points, 3) or (num_points, 2)."

    if persistence_diagram.shape[1] == 2:
        return OneHotEncodedPersistenceDiagram(
            torch.stack(
                (
                    torch.tensor(persistence_diagram),
                    torch.ones(persistence_diagram.shape[0]),
                ),
                dim=1,
            )
        )
    else:
        homology_types: Set[int] = set([int(i) for i in persistence_diagram[:, 2]])
        type_to_one_hot_encoding: Dict[int, int] = {
            i: j for j, i in enumerate(homology_types)
        }
        one_hot_encoding: Tensor = torch.zeros(
            persistence_diagram.shape[0], len(homology_types)
        )
        # TODO: Fill the one-hot encoding in a vectorized manner.
        for i, j in enumerate(persistence_diagram[:, 2]):
            one_hot_encoding[i, type_to_one_hot_encoding[int(j)]] = 1
        birth_death_diagram: Tensor = torch.tensor(persistence_diagram[:, :2])

        return OneHotEncodedPersistenceDiagram(
            torch.cat((birth_death_diagram, one_hot_encoding), dim=1).float()
        )


def get_one_hot_encoded_persistence_diagram_from_gudhi_extended(
    diagram: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
) -> OneHotEncodedPersistenceDiagram:
    """Convert an extended persistence diagram of a single graph to an
    array with one-hot encoded homology type.
    Args:
        diagram (Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]):
            The diagram of an extended persistence of a single graph.

    Returns:
        np.ndarray:
            The diagram in one-hot encoded homology type of size
            (num_points, 6).
    """
    # Get the length of each array
    lengths = [len(array) for array in diagram]

    if lengths == [0, 0, 0, 0]:
        return OneHotEncodedPersistenceDiagram(torch.zeros((0, 6)))

    # One-hot encode the homology type
    homology_type: np.ndarray = np.array(
        flatten_list_of_lists([[i] * lengths[i] for i in range(4)])
    )
    homology_type_one_hot = np.zeros((sum(lengths), 4))
    homology_type_one_hot[np.arange(sum(lengths)), homology_type] = 1

    # Concatenate the arrays
    diagram_one_hot = np.concatenate([sub_diagram for sub_diagram in diagram], axis=0)
    diagram_one_hot = np.concatenate([diagram_one_hot, homology_type_one_hot], axis=1)

    return OneHotEncodedPersistenceDiagram.from_numpy(diagram_one_hot)  # type: ignore


def _plot_diagram(
    diagram, names: List[str], homology_dimensions=None, plotly_params=None
):
    """Plot a single persistence diagram.

    Args:
        diagram, ndarray of shape (n_points, 3):
            The persistence diagram to plot, where the third dimension along axis 1
            contains homology dimensions, and the first two contain (birth, death)
            pairs to be used as coordinates in the two-dimensional plot.
        homology_dimensions, list of int or None, optional, default: ``None``:
            Homology dimensions which will appear on the plot. If ``None``, all
            homology dimensions which appear in `diagram` will be plotted.
        plotly_params, dict or None, optional, default: ``None``:
            Custom parameters to configure the plotly figure. Allowed keys are
            ``"traces"`` and ``"layout"``, and the corresponding values should be
            dictionaries containing keyword arguments as would be fed to the
            :meth:`update_traces` and :meth:`update_layout` methods of
            :class:`plotly.graph_objects.Figure`.
    Returns:
        fig :
            :class:`plotly.graph_objects.Figure` object;
            Figure representing the persistence diagram.
    """
    # TODO: increase the marker size
    if homology_dimensions is None:
        homology_dimensions = np.unique(diagram[:, 2])

    diagram = diagram[diagram[:, 0] != diagram[:, 1]]
    diagram_no_dims = diagram[:, :2]
    posinfinite_mask = np.isposinf(diagram_no_dims)
    neginfinite_mask = np.isneginf(diagram_no_dims)
    max_val = np.max(np.where(posinfinite_mask, -np.inf, diagram_no_dims))
    min_val = np.min(np.where(neginfinite_mask, np.inf, diagram_no_dims))
    parameter_range = max_val - min_val
    extra_space_factor = 0.02
    has_posinfinite_death = np.any(posinfinite_mask[:, 1])  # type: ignore
    if has_posinfinite_death:
        posinfinity_val = max_val + 0.1 * parameter_range
        extra_space_factor += 0.1
    extra_space = extra_space_factor * parameter_range
    min_val_display = min_val - extra_space
    max_val_display = max_val + extra_space

    fig = gobj.Figure()
    fig.add_trace(
        gobj.Scatter(
            x=[min_val_display, max_val_display],
            y=[min_val_display, max_val_display],
            mode="lines",
            line={"dash": "dash", "width": 1, "color": "black"},
            showlegend=False,
            hoverinfo="none",
        )
    )

    for dim in homology_dimensions:
        name = names[int(dim)]  # type: ignore
        subdiagram = diagram[diagram[:, 2] == dim]
        unique, inverse, counts = np.unique(  # type: ignore
            subdiagram, axis=0, return_inverse=True, return_counts=True
        )
        hovertext = [
            f"{tuple(unique[unique_row_index][:2])}"
            + (
                f", multiplicity: {counts[unique_row_index]}"
                if counts[unique_row_index] > 1
                else ""
            )
            for unique_row_index in inverse
        ]
        y = subdiagram[:, 1]
        if has_posinfinite_death:
            y[np.isposinf(y)] = posinfinity_val  # type: ignore
        fig.add_trace(
            gobj.Scatter(
                x=subdiagram[:, 0],
                y=y,
                mode="markers",
                hoverinfo="text",
                hovertext=hovertext,
                name=name,
            )
        )

    fig.update_layout(
        width=500,
        height=500,
        xaxis1={
            "title": "Birth",
            "side": "bottom",
            "type": "linear",
            "range": [min_val_display, max_val_display],
            "autorange": False,
            "ticks": "outside",
            "showline": True,
            "zeroline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": False,
            "showexponent": "all",
            "exponentformat": "e",
        },
        yaxis1={
            "title": "Death",
            "side": "left",
            "type": "linear",
            "range": [min_val_display, max_val_display],
            "autorange": False,
            "scaleanchor": "x",
            "scaleratio": 1,
            "ticks": "outside",
            "showline": True,
            "zeroline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": False,
            "showexponent": "all",
            "exponentformat": "e",
        },
        plot_bgcolor="white",
    )

    # Add a horizontal dashed line for points with infinite death
    if has_posinfinite_death:
        fig.add_trace(
            gobj.Scatter(
                x=[min_val_display, max_val_display],
                y=[posinfinity_val, posinfinity_val],  # type: ignore
                mode="lines",
                line={"dash": "dash", "width": 0.5, "color": "black"},
                showlegend=True,
                name="\u221E",
                hoverinfo="none",
            )
        )

    # Update traces and layout according to user input
    if plotly_params:
        fig.update_traces(plotly_params.get("traces", None))
        fig.update_layout(plotly_params.get("layout", None))

    return fig



from collections import Counter, OrderedDict
from typing import Callable, Tuple, Optional, List, Dict, Union

import torch
from torchtext.data.utils import get_tokenizer  # type: ignore
from torchtext.vocab import vocab
from gdeep.utility import DEVICE

from ..abstract_preprocessing import AbstractPreprocessing
from .._utils import MissingVocabularyError

# type definition
from gdeep.utility.custom_types import Tensor


class TokenizerTranslation(
    AbstractPreprocessing[Tuple[str, str], Tuple[Union[Tensor, List[Tensor]], Tensor]]
):
    """Class to preprocess text dataloaders for translation
    tasks. The Dataset type is supposed to be ``(string, string)``.
    The padding item is supposed to be of index 0.

        Args:
            vocabulary :
                the vocabulary of the source text;
                it can be built automatically or it can be
                given.
            vocabulary_target :
                the vocubulary of the target text;
                it can be built automatically or it can be
                given.
            tokenizer:
                the tokenizer of the source text
            tokenizer_target:
                the tokenizer of the target text

    Examples::

        from gdeep.data import DatasetBuilder
        from gdeep.data import TransformingDataset
        from gdeep.data.preprocessors import TokenizerTranslation

        db = DatasetBuilder(name="Multi30k", convert_to_map_dataset=True)
        ds_tr, ds_val, _ = db.build()

        textds = TransformingDataset(ds_tr,
            TokenizerTranslation())

    """

    if_fitted: bool
    vocabulary: Optional[Dict[str, int]]
    vocabulary_target: Optional[Dict[str, int]]
    tokenizer: Optional[Callable[[str], List[str]]]
    tokenizer_target: Optional[Callable[[str], List[str]]]
    ordered_dict: Dict[str, int]
    ordered_dict_target: Dict[str, int]

    def __init__(
        self,
        vocabulary: Optional[Dict[str, int]] = None,
        vocabulary_target: Optional[Dict[str, int]] = None,
        tokenizer: Optional[Callable[[str], List[str]]] = None,
        tokenizer_target: Optional[Callable[[str], List[str]]] = None,
    ):

        if tokenizer is None:
            self.tokenizer = get_tokenizer("basic_english")
        else:
            self.tokenizer = tokenizer
        if tokenizer_target is None:
            self.tokenizer_target = get_tokenizer("basic_english")
        else:
            self.tokenizer_target = tokenizer_target
        self.vocabulary = vocabulary
        self.vocabulary_target = vocabulary_target
        self.max_length_source: int = 0
        self.max_length_target: int = 0
        self.is_fitted = False

    def fit_to_dataset(self, data):
        self.counter = Counter()
        self.counter_target = Counter()
        self.ordered_dict = OrderedDict()
        self.ordered_dict_target = OrderedDict()
        assert self.tokenizer is not None, "Tokenizer is None"
        assert self.tokenizer_target is not None, "Target Tokenizer is None"
        for text in data:
            self.counter.update(self.tokenizer(text[0]))
            self.counter_target.update(self.tokenizer_target(text[1]))
            self.max_length_source = max(
                self.max_length_source, len(self.tokenizer(text[0]))
            )
            self.max_length_target = max(
                self.max_length_target, len(self.tokenizer_target(text[1]))
            )
        # self.vocabulary = Vocab(counter, min_freq=1)
        if self.vocabulary is None:
            self.ordered_dict = OrderedDict(
                sorted(self.counter.items(), key=lambda x: x[1], reverse=True)
            )
            self.vocabulary = vocab(self.ordered_dict)  # type: ignore
            unk_token = "<unk>"  # type: ignore
            if unk_token not in self.vocabulary:
                self.vocabulary.insert_token(unk_token, 0)
            self.vocabulary.set_default_index(self.vocabulary[unk_token])
        if self.vocabulary_target is None:
            self.ordered_dict_target = OrderedDict(
                sorted(self.counter_target.items(), key=lambda x: x[1], reverse=True)
            )
            self.vocabulary_target = vocab(self.ordered_dict_target)  # type: ignore
            unk_token = "<unk>"  # type: ignore
            if unk_token not in self.vocabulary_target:
                self.vocabulary_target.insert_token(unk_token, 0)
            self.vocabulary_target.set_default_index(self.vocabulary_target[unk_token])
            # self.vocabulary_target = Vocab(self.counter_target)
        self.is_fitted = True
        # self.save_pretrained(".")

    def __call__(
        self, datum: Tuple[str, str]
    ) -> Tuple[Union[List[Tensor], Tensor], Tensor]:
        """This method is applied to each item of
        the dataset and
        transforms it following the rule described
        in this method

        Args:
            datum (torch.tensor):
                a single datum
        """

        # if not self.is_fitted:
        #    self.load_pretrained(".")
        if self.vocabulary:
            text_pipeline: Callable[[str], List[int]] = lambda x: [
                self.vocabulary[token] for token in self.tokenizer(x)  # type: ignore
            ]  # type: ignore
        else:
            raise MissingVocabularyError(
                "Please fit this preprocessor to initialise the vocabulary"
            )

        if self.vocabulary_target:
            text_pipeline_target: Callable[[str], List[int]] = lambda xx: [
                self.vocabulary_target[token] for token in self.tokenizer_target(xx)  # type: ignore
            ]  # type: ignore
        else:
            raise MissingVocabularyError(
                "Please fit this preprocessor to initialise the vocabulary"
            )

        pad_item: int = 0
        pad_item_target: int = 0

        processed_text = torch.tensor(text_pipeline(datum[0]), dtype=torch.long).to(
            DEVICE
        )
        processed_text_target = torch.tensor(
            text_pipeline_target(datum[1]), dtype=torch.long
        ).to(DEVICE)
        # convert to tensors (padded)
        out_source = torch.cat(
            [
                processed_text,
                pad_item
                * torch.ones(self.max_length_source - processed_text.shape[0]).to(
                    DEVICE
                ),
            ]
        ).to(torch.long)
        out_target = torch.cat(
            [
                processed_text_target,
                pad_item_target
                * torch.ones(
                    self.max_length_target - processed_text_target.shape[0]
                ).to(DEVICE),
            ]
        ).to(torch.long)
        x = [out_source.to(torch.long), out_target.to(torch.long)]
        y = out_target.to(torch.long)  # .clone().detach()
        return x, y



from .persistence_activations import (
    persistence_diagrams_of_activations,
    _simplified_persistence_of_activations,
)
from .plot_compactification import Compactification
from .utils import plotly2tensor, png2tensor
from .visualiser import Visualiser


__all__ = [
    "Compactification",
    "persistence_diagrams_of_activations",
    "plotly2tensor",
    "png2tensor",
    "Visualiser",
]



import os
from typing import List, Tuple
import ssl
import urllib
import zipfile
from collections import defaultdict

from torch.utils.data import Dataset
import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.sparse import coo_matrix

from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from gdeep.utility.extended_persistence import graph_extended_persistence_hks
from gdeep.utility.constants import DEFAULT_GRAPH_DIR


PD = OneHotEncodedPersistenceDiagram


class PersistenceDiagramFromGraphBuilder:
    """
    This class is used to load the persistence diagrams of the graphs in a
    dataset. All graph datasets available at https://chrsmrrs.github.io/datasets/docs/datasets/
    are supported.
    """

    url: str = "https://www.chrsmrrs.com/graphkerneldatasets"

    def __init__(
        self,
        dataset_name: str,
        diffusion_parameter: float,
        root: str = DEFAULT_GRAPH_DIR,
    ):
        """
        Initialize the dataset.

        Args:
            dataset_name:
                The name of the graph dataset to load, e.g. "MUTAG".
            diffusion_parameter:
                The diffusion parameter of the heat kernel
                signature. These are usually chosen to be as {0.1, 1.0, 10.0}.
            root:
                The directory where to store the dataset
        """
        self.dataset_name: str = dataset_name
        self.diffusion_parameter: float = diffusion_parameter
        self.num_homology_types: int = 4
        self.root: str = root
        self.output_dir: str = os.path.join(
            root,
            dataset_name + "_" + str(diffusion_parameter) + "_extended_persistence",
        )

    def create(self) -> None:
        # Check if the dataset exists in the specified directory
        if not os.path.exists(self.output_dir):
            print(f"Dataset {self.dataset_name} does not exist!")
            self._preprocess()
        else:
            print(
                f"Dataset {self.dataset_name} already exists!"
                " Skipping: dataset will not be created."
            )

    def __repr__(self) -> str:
        """
        Return a string representation of the dataset.

        Returns:
            A string representation of the dataset.
        """
        return (
            f"{self.__class__.__name__}(dataset_name={self.dataset_name}, "
            f"diffusion_parameter={self.diffusion_parameter}, "
            f"root={self.root})"
        )

    def _preprocess(self) -> None:
        """
        Preprocess the dataset and save the persistence diagrams and the labels
        in the output directory.
        The persistence diagrams are computed using the heat kernel signature
        method and then each diagram is saved in a separate npy file in the
        diagrams subdirectory of the output directory.
        The labels are saved in a csv file in the output directory.

        """
        # Load the dataset
        self.graph_dataset = GraphDataset(
            root=DEFAULT_GRAPH_DIR, dataset_name=self.dataset_name, url=self.url
        )

        # Create the directory for PDs if it does not exist
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
            os.makedirs(os.path.join(self.output_dir, "diagrams"))
        else:
            # This should not be reached
            raise ValueError("Output directory already exists!")

        num_graphs = len(self.graph_dataset)

        labels: List[Tuple[int, int]] = []

        print("Computing the persistence diagrams...")
        for graph_idx, graph in tqdm(
            enumerate(self.graph_dataset), total=num_graphs  # type: ignore
        ):

            # Get the adjacency matrix
            adj_mat: np.ndarray = graph[0]

            # Compute the extended persistence
            persistence_diagram_one_hot = graph_extended_persistence_hks(
                adj_mat, diffusion_parameter=self.diffusion_parameter
            )
            # Sort the diagram by the persistence lifetime, i.e. the second
            # column minus the first column

            # Save the persistence diagram in a file
            persistence_diagram_one_hot.save(
                os.path.join(self.output_dir, "diagrams", f"{graph_idx}.npy")
            )

            # Save the label
            labels.append((graph_idx, graph[1]))

        # Save the labels in a csv file
        pd.DataFrame(labels, columns=["graph_idx", "label"]).to_csv(
            os.path.join(self.output_dir, "labels.csv"), index=False
        )


class GraphDataset(Dataset):
    """This class i a light weight built-i class to retrieve
    data from the TUDatasets https://chrsmrrs.github.io/datasets/docs/datasets/.

    Args:
        dataset_name:
            name of the dataset appearing in the TUDataset webpage
        root:
            the root folder where to store the .zip file
        url:
            the url at which to fetch the dataset, most likely of the form
            https://www.chrsmrrs.com/graphkerneldatasets/<datasetname.zip>
    """

    def __init__(self, dataset_name: str, root: str, url: str):
        self.dataset_name = dataset_name
        self.root = root
        self.url = url
        self.dataset: List[
            Tuple[np.ndarray, int]
        ] = []  # where to store the adj matrices in memory
        self._build_dataset()

    def _download_url(self, url: str, folder: str) -> str:
        """private method to download the .zip file of
        the dataset"""
        context = ssl._create_unverified_context()  # noqa
        data = urllib.request.urlopen(url, context=context)  # type: ignore # noqa
        if not os.path.exists(folder):
            os.makedirs(folder)
        path_to_zip = os.path.join(folder, self.dataset_name + ".zip")
        with open(path_to_zip, "wb") as f:
            # workaround for https://bugs.python.org/issue42853
            while True:
                chunk = data.read(10 * 1024 * 1024)
                if not chunk:
                    break
                f.write(chunk)

        return path_to_zip

    @staticmethod
    def _extract_zip(path_to_zip: str, folder: str) -> None:
        """Private method to extract the zip file of the dataset"""
        with zipfile.ZipFile(path_to_zip, "r") as f:
            f.extractall(folder)

    def _download(self) -> str:
        """private method to download and extract the dataset"""
        folder = os.path.join(self.root, self.dataset_name)
        path_to_zip = self._download_url(f"{self.url}/{self.dataset_name}.zip", folder)
        self._extract_zip(path_to_zip, folder)
        os.unlink(path_to_zip)
        return path_to_zip[:-4]
        # shutil.rmtree(self.raw_dir)
        # os.rename(os.path.join(folder, self.name), self.raw_dir)

    def _build_dataset(self) -> None:
        """private method to put the dataset in memory. The data is
        stored in ``self.dataset``"""
        path = self._download()  # location of the files

        idx_file = os.path.join(path, self.dataset_name + "_graph_indicator.txt")
        indices = np.loadtxt(idx_file, delimiter=",", dtype=np.int32).T - 1  # type: ignore

        graph_labels = os.path.join(path, self.dataset_name + "_graph_labels.txt")
        labels = (np.loadtxt(graph_labels, delimiter=",", dtype=np.int32).T + 1) // 2  # type: ignore

        adj_file = os.path.join(path, self.dataset_name + "_A.txt")
        sparse_array = np.loadtxt(adj_file, delimiter=",", dtype=np.int32).T - 1  # type: ignore

        rows: defaultdict = defaultdict(list)
        cols: defaultdict = defaultdict(list)

        for node_id in sparse_array[0]:
            rows[indices[node_id]] += [node_id]

        for node_id in sparse_array[1]:
            cols[indices[node_id]] += [node_id]

        for idx, _ in enumerate(rows):
            row = np.array(rows[idx])  # np.array([0, 3, 1, 0])
            col = np.array(cols[idx])  # np.array([0, 3, 1, 2])
            data = np.ones(len(row))  # np.array([4, 5, 7, 9])
            assert len(col) == len(row)
            shape_x = max(max(row) - min(row), max(col) - min(col)) + 1
            offset = min(min(row), min(col))
            # print(row, col, offset, shape_x)
            adj_mat = coo_matrix(
                (data, (row - offset, col - offset)), shape=(shape_x, shape_x)
            ).toarray()
            # print(adj_mat, idx)
            self.dataset.append((adj_mat, labels[idx]))

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]



from abc import ABC, abstractmethod
from typing import List, Any, Union

import torch
from torch import nn

from gdeep.utility.custom_types import Tensor


class SaveOutput(ABC):
    """General class for saving outputs of
    neural networks.
    Outputs will be stored in a list 'outputs'
    """

    outputs: List[Any]

    @abstractmethod
    def __call__(
        self,
        module: nn.Module,
        module_in: Tensor,
        module_out: Tensor,
    ) -> None:
        pass

    def clear(self) -> None:
        self.outputs = []

    def get_outputs(self) -> List[Any]:
        return self.outputs


class SaveNodeOutput(SaveOutput):
    """Class for saving activations of a node in
    a neural network.

    Args:
        entry:
            select the entry row in the output
            of the layer, i.e. the node output
    """

    def __init__(self, entry: int = 0):
        self.outputs = []
        self.entry = entry

    def __call__(self, module: nn.Module, module_in: Tensor, module_out: Tensor):
        self.outputs.append(module_out[:, self.entry].detach())


class SaveLayerOutput(SaveOutput):
    def __init__(self) -> None:
        self.outputs = []

    def __call__(
        self,
        module: nn.Module,
        module_in: Tensor,
        module_out: Tensor,
    ) -> None:
        """This function stores the layer activations

        Args:
            module :
                the neural network model (nn.Module)
            module_in:
                the input tensor of the module
            module_out:
                the output tensor of the module
        """
        if isinstance(module_out, tuple):
            if hasattr(module_out[0], "detach"):
                self.outputs.append(module_out[0].detach())
            else:
                self.outputs.append(module_out[0])  # .detach())
        else:
            if hasattr(module_out, "detach"):
                self.outputs.append(module_out.detach())
            else:
                self.outputs.append(module_out)  # .detach())



from typing import Callable, Optional, Tuple
import torch.nn as nn
import torch.nn.functional as F

from gdeep.utility.custom_types import Tensor


# build a custom network class to easily do experiments
class FFNet(nn.Module):
    """Custom network class to easily do experiments

    It is a simple feed-forward network with a variable number of layers and
    neurons. The activation function is ReLU by default, but it can be changed
    by passing a different function as argument.
    The last layer is not activated.

    Args:
        arch (Tuple[int, ...], optional):
            The architecture of the network.
            Tuple containing the dimension of the layers
            inside your network. The default is (2, 3, 3, 2).
            The depth of the network is inferred from the length of the tuple.

        activation (Optional[Callable], optional):
            The activation function.
            Defaults to None. If None, ReLU is used.
            All the layers are activated with the same function. The last layer
            is not activated.
    """

    def __init__(
        self,
        arch: Tuple[int, ...] = (2, 3, 3, 2),
        activation: Optional[Callable] = None,
    ):
        super(FFNet, self).__init__()
        if activation is None:
            self.activation = F.relu
        else:
            self.activation = activation
        self.linears = nn.ModuleList(
            [nn.Linear(arch[i], arch[i + 1]) for i in range(len(arch) - 1)]
        )

    def forward(self, x: Tensor) -> Tensor:
        for i, l in enumerate(self.linears):
            if i < len(self.linears) - 1:
                x = self.activation(l(x))
            else:
                x = l(x)
        return x



from typing import Optional

import torch
from gdeep.utility.enum_types import (
    ActivationFunction,
    AttentionType,
    LayerNormStyle,
    PoolerType,
)
from torch.nn import Module

from .persformer import Persformer
from .persformer_config import PersformerConfig

from gdeep.utility.custom_types import Tensor


class PersformerWrapper(Module):
    """The wrapper for persformer to allow compatibility
    with the HPO classes.
    """

    config: PersformerConfig
    model: Module

    def __init__(self, **kwargs):
        super().__init__()
        self.config = PersformerConfig(**kwargs)
        self.model = Persformer(self.config)

    def forward(self, input: Tensor, attention_mask: Optional[Tensor] = None):
        return self.model(input, attention_mask)



from typing import List, Tuple, Union, Optional

# deep learning
import torch
import numpy as np
from torch import nn

# plot
import plotly.express as px
import pandas as pd
from plotly.graph_objs._figure import Figure  # noqa

# ML
from sklearn.metrics import pairwise_distances

from ..analysis.decision_boundary import (
    UniformlySampledPoint,
    GradientFlowDecisionBoundaryCalculator,
)

from gdeep.utility.custom_types import Tensor

from gdeep.utility.custom_types import Array


class Compactification:
    """Class to compactify the feature space and get
    better topological insights.

    Args:
        precision :
            A tolerance on how close to the decision
            boundary the points shall be.
        n_features :
            number of input features
        n_samples:
            number of samples per chart
        n_epochs:
            number of epochs needed to push points to the
            decision boundary
        boundary_tuple :
            list of pairs (left,right).
            This list defines the boundaries in each coordinates
        neural_net :
            the trained network of which to compute
            the boundary
    """

    patches: List
    list_of_pts_in_patches: List
    label_final: List

    def __init__(
        self,
        neural_net: nn.Module,
        precision: float = 0.4,
        n_features: Optional[int] = None,
        n_samples: int = 1000,
        epsilon: float = 0.05,
        n_epochs: int = 5000,
        boundary_tuple: Optional[List[Tuple[float, float]]] = None,
    ) -> None:

        if n_features is None:
            raise ValueError("Need to specify the number of features.")
        else:
            self.n_features = n_features
        self.neural_net = neural_net
        if boundary_tuple:
            self.boundary_tuple = boundary_tuple
        else:
            self.boundary_tuple = [(-1.0, 1.0) for _ in range(self.n_features)]

        self.precision = precision
        self.n_samples = n_samples
        self.epsilon = epsilon
        self.n_epochs = n_epochs
        self.check_compute_charts = False

    @staticmethod
    def _transition_to_patch(sample_points_tensor: Tensor, i: int) -> Tensor:
        stacking_list = []
        if i == -1:
            points = sample_points_tensor
            return points
        else:
            for dim in range(sample_points_tensor.shape[1]):
                if dim == i:
                    stacking_list.append(1 / sample_points_tensor[:, dim])
                else:
                    stacking_list.append(
                        sample_points_tensor[:, dim] / sample_points_tensor[:, i]
                    )
            return torch.stack(stacking_list, dim=1)

    def _compute_charts(self) -> List[Tensor]:
        # compute boundary
        self.patches = []

        # loop over patches
        for i in range(-1, self.n_features):
            # print(i)
            # sample pts uniformly to 0-th patch
            sample_points = UniformlySampledPoint(
                self.boundary_tuple, n_samples=self.n_samples
            )
            plot_points_tensor = torch.from_numpy(sample_points()).float()
            # move to i-th patch
            sample_points_tensor = self._transition_to_patch(plot_points_tensor, i)

            # Using new gradient flow implementation
            gf = GradientFlowDecisionBoundaryCalculator(
                model=self.neural_net,
                initial_points=sample_points_tensor,
                optimizer=lambda params: torch.optim.Adam(params),
            )
            gf.step(number_of_steps=self.n_epochs)
            res = gf.get_filtered_decision_boundary(delta=self.precision).detach().cpu()
            # back to 0-th patch
            plot_points_tensor = self._transition_to_patch(res, i)
            self.patches.append(plot_points_tensor)
        self.check_compute_charts = True
        return self.patches

    def create_final_distance_matrix(self) -> Tuple[Array, List]:
        """This is the main function to call without arguments"""
        if self.check_compute_charts is False:
            self._compute_charts()
        self.list_of_pts_in_patches = []
        for i in range(len(self.patches)):
            list_of_pts_per_patch = []
            for j, pts in enumerate(self.patches):
                if j == i:
                    list_of_pts_per_patch.append(pts)
                else:
                    list_of_pts_per_patch.append(
                        self._transition_to_patch(
                            self._transition_to_patch(pts, j - 1), i - 1
                        )
                    )
            self.list_of_pts_in_patches.append(np.concatenate(list_of_pts_per_patch))

        list_of_distances = []
        for array in self.list_of_pts_in_patches:
            list_of_distances.append(pairwise_distances(array))

        d_patch = np.stack(list_of_distances, axis=2)
        d_final = np.min(d_patch, axis=2)
        self.label_final = []
        for j, patch in enumerate(self.patches):
            self.label_final += list(j * np.ones(len(patch)))
        return d_final, self.label_final

    def plot_chart(self, i: int) -> Figure:
        """This functions plots the points in each chart.

        Args:
            i :
                the chart index, from ``-1`` to ``n_features``
        """
        df_plot = pd.DataFrame(
            self.list_of_pts_in_patches[i],
            columns=[
                "x" + str(j) for j in range(self.list_of_pts_in_patches[i].shape[1])
            ],
        )
        df_plot["label"] = self.label_final
        if self.list_of_pts_in_patches[i].shape[1] == 2:
            fig = px.scatter(df_plot, x="x0", y="x1", color="label")

        elif self.list_of_pts_in_patches[i].shape[1] >= 3:
            fig = px.scatter_3d(df_plot, x="x0", y="x1", z="x2", color="label")
        else:
            raise ValueError("Need at least 2 features to plot.")
        return fig



from gdeep.data.datasets import (
    ImageClassificationFromFiles,
    DataLoaderBuilder,
    FromArray,
    DlBuilderFromDataCloud,
)

from gdeep.data.preprocessors import ToTensorImage
import os
from os.path import join
import numpy as np
import torch
import logging
from google.auth.exceptions import DefaultCredentialsError  # type: ignore

LOGGER = logging.getLogger(__name__)


def test_array():
    """test class DatasetFromArray"""
    X = np.random.rand(10, 4)  # type: ignore
    y = np.random.randint(3, size=10)
    ds = FromArray(X, y)
    dl, *_ = DataLoaderBuilder((ds,)).build(({"batch_size": 1},))  # type: ignore
    item = next(iter(dl))
    assert torch.norm(item[0] - torch.tensor(X[0])) < 1e-6
    assert item[1] == torch.tensor(y[0])
    assert len(dl.dataset) == 10  # type: ignore


def test_array_tensor():
    """test class DatasetFromArray"""
    X = torch.rand(10, 4)
    y = torch.randint(3, size=(10,))
    ds = FromArray(X, y)
    dl, *_ = DataLoaderBuilder((ds,)).build()  # type: ignore
    item = next(iter(dl))
    assert torch.norm(item[0] - X[0]) < 1e-6
    assert item[1] == y[0]



from typing import Optional

import torch
from torch.nn import Dropout, Linear, Module, ModuleList, Sequential

from .attention_factory import AttentionFactory
from .persformer_block import PersformerBlock
from .persformer_config import PersformerConfig
from .utility import get_activation_function, get_pooling_layer

# Type aliases
from gdeep.utility.custom_types import Tensor


class Persformer(Module):
    """Persformer model as described in the paper: https://arxiv.org/abs/2112.15210

    Examples::

        from gdeep.topology_layers import PersformerConfig, PersformerModel
        # Initialize the configuration object
        config = PersformerConfig()
        # Initialize the model
        model = Persformer(config)

    """

    config: PersformerConfig
    embedding_layer: Module
    persformer_blocks: ModuleList
    classifier_layer: Module
    pooling_layer: Module

    def __init__(self, config: PersformerConfig):
        super().__init__()
        self.config = config
        self.build_model()

    def build_model(self):
        """
        Build the model.
        """
        self.embedding_layer = self._get_embedding_layer()
        self.persformer_blocks = ModuleList(
            [
                self._get_persformer_block()
                for _ in range(self.config.num_attention_layers)
            ]
        )
        self.pooling_layer = self._get_pooling_layer()
        self.classifier_layer = self._get_classifier_layer()

    def _get_embedding_layer(self) -> Module:
        return Sequential(
            Linear(self.config.input_size, self.config.hidden_size),
            get_activation_function(self.config.hidden_act),
        )

    def _get_classifier_layer(self) -> Module:
        return Sequential(
            Linear(self.config.hidden_size, self.config.hidden_size),
            get_activation_function(self.config.hidden_act),
            Dropout(self.config.classifier_dropout_prob),
            Linear(self.config.hidden_size, self.config.output_size),
        )

    def _get_persformer_block(self) -> Module:
        return PersformerBlock(self.config)  # type: ignore

    def _get_pooling_layer(self) -> Module:
        return get_pooling_layer(self.config)

    def forward(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model.

        Args:
            input_batch: The input batch. Of shape (batch_size, sequence_length, 2 + num_homology_types)
            attention_mask: The attention mask. Of shape (batch_size, sequence_length)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        # Initialize the output tensor
        output = input_batch
        # Apply the embedding layer
        output = self.embedding_layer(output)
        # Apply the attention layers
        for persformer_block in self.persformer_blocks:
            output = persformer_block(output, attention_mask)
        output = self.pooling_layer(output, attention_mask)
        # Apply the classifier layer
        output = self.classifier_layer(output)
        return output



from typing import Optional

import torch
from gdeep.utility.enum_types import LayerNormStyle
from torch.nn import LayerNorm, Module, ModuleList

from .persformer_config import PersformerConfig
from .utility import get_attention_layer, get_feed_forward_layer

# Type aliases
from gdeep.utility.custom_types import Tensor


class PersformerBlock(Module):
    """
    A persformer block.
    """

    config: PersformerConfig
    attention_layer: Module
    feed_forward_layer: Module
    dropout_layer: Module
    layer_norms: Optional[ModuleList]

    def __init__(self, config: PersformerConfig):
        super().__init__()
        self.config = config
        self.build_model()

    def build_model(self):
        """
        Build the model.
        """
        self.attention_layer = get_attention_layer(self.config)
        if not self.config.use_attention_only:
            self.feed_forward_layer = get_feed_forward_layer(self.config)

        if (
            self.config.layer_norm_style == LayerNormStyle.PRE_LAYER_NORMALIZATION
            or self.config.layer_norm_style == LayerNormStyle.POST_LAYER_NORMALIZATION
        ):
            self.layer_norms = ModuleList(
                [
                    LayerNorm(self.config.hidden_size, eps=self.config.layer_norm_eps)
                    for _ in range(2)
                ]
            )

    def forward(
        self,  # type: ignore
        input_batch: Tensor,
        attention_mask: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Forward pass of the model. We implement different layer normalizations in the forward pass,
        see the paper https://arxiv.org/pdf/2002.04745.pdf for details.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, 2 + num_homology_types)
            attention_mask:
                The attention mask. Of shape (batch_size, sequence_length)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        if self.config.layer_norm_style == LayerNormStyle.NO_LAYER_NORMALIZATION:
            return self._forward_no_layer_norm(input_batch, attention_mask)
        elif self.config.layer_norm_style == LayerNormStyle.PRE_LAYER_NORMALIZATION:
            return self._forward_pre_layer_norm(input_batch, attention_mask)
        elif self.config.layer_norm_style == LayerNormStyle.POST_LAYER_NORMALIZATION:
            return self._forward_post_layer_norm(input_batch, attention_mask)
        else:
            raise ValueError(f"Unknown layer norm style {self.config.layer_norm_style}")

    def _forward_no_layer_norm(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model without layer normalization.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, 2 + num_homology_types)
            attention_mask:
                The attention mask. Of shape (batch_size, sequence_length)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        output = self.attention_layer(input_batch, attention_mask) + input_batch  # type: ignore
        if not self.config.use_attention_only:
            output = self.feed_forward_layer(output) + output
        return output

    def _forward_pre_layer_norm(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model with pre-layer normalization.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, 2 + num_homology_types)
            attention_mask:
                The attention mask. Of shape (batch_size, sequence_length)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        assert self.layer_norms is not None
        normalized = self.layer_norms[0](input_batch)
        output = self.attention_layer(normalized, attention_mask) + input_batch
        if not self.config.use_attention_only:
            normalized = self.layer_norms[1](output)
            output = self.feed_forward_layer(normalized) + output
        return output

    def _forward_post_layer_norm(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model with post-layer normalization.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, 2 + num_homology_types)
            attention_mask:
                The attention mask. Of shape (batch_size, sequence_length)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        assert self.layer_norms is not None
        output = self.attention_layer(input_batch, attention_mask) + input_batch
        output = self.layer_norms[0](output)
        if not self.config.use_attention_only:
            output = self.feed_forward_layer(output) + output
            output = self.layer_norms[1](output)
        return output



from .categorical_data import CategoricalDataCloud
from .tori import Rotation, ToriDataset
from .dataset_cloud import DatasetCloud
from ._data_cloud import _DataCloud
from .build_datasets import DatasetBuilder, get_dataset
from .base_dataloaders import DataLoaderBuilder, AbstractDataLoaderBuilder
from .dataset_for_image import ImageClassificationFromFiles
from .dataset_form_array import FromArray
from .dataloader_cloud import DlBuilderFromDataCloud
from .parallel_orbit import (
    generate_orbit_parallel,
    create_pd_orbits,
    OrbitsGenerator,
    DataLoaderKwargs,
)
from .persistence_diagrams_from_files import PersistenceDiagramFromFiles


__all__ = [
    "Rotation",
    "CategoricalDataCloud",
    "AbstractDataLoaderBuilder",
    "ToriDataset",
    "DataLoaderBuilder",
    "DatasetBuilder",
    "generate_orbit_parallel",
    "create_pd_orbits",
    "OrbitsGenerator",
    "DataLoaderKwargs",
    "get_dataset",
    "ImageClassificationFromFiles",
    "FromArray",
    "DatasetCloud",
    "DlBuilderFromDataCloud",
    "PersistenceDiagramFromFiles",
]



from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from gdeep.utility.custom_types import Tensor


class PeriodicNeuralNetwork(nn.Module):
    """Makes a periodic `nn.Module` of `nn`. `boundary_list` specifies
    the elementary region the periodic neural network is supported on.
    This creates neural networks
    factorized by the orbits of a group effect,
    which is given by reflection at the
    edges of the elementary domain.
    This class can be interpreted as an adapter for periodic neural networks.

    Args:
        nn :
            the standard torch module, your network
        boundary_list :
            list of pairs of floats, each
            defining the boundaries of a hypercube
    """

    def __init__(self, nn: nn.Module, boundary_list: List[Tuple[float, float]]) -> None:
        super().__init__()
        self.nn = nn
        self.interval_length = torch.tensor([[b - a for a, b in boundary_list]])
        self.left_interval_bound = torch.tensor([[a for a, b in boundary_list]])

    def forward(self, x_cont: Tensor) -> Tensor:
        x_cont = (
            torch.abs(
                torch.remainder(
                    x_cont - self.left_interval_bound, 2.0 * self.interval_length
                )
                - self.interval_length
            )
            + self.left_interval_bound
        )
        x_cont = self.nn.forward(x_cont)
        return x_cont



from .persformer_config import PersformerConfig

from .persformer import Persformer

from .persformer_wrapper import PersformerWrapper

from . import attention, pooling_layers

__all__ = [
    "PersformerConfig",
    "Persformer",
    "PersformerWrapper",
    "attention",
    "pooling_layers",
]



"""Testing for decision_boundary_calculator."""
# License: GNU AGPLv3
from .. import GradientFlowDecisionBoundaryCalculator
from gdeep.data import *
from ....models import FFNet
import torch

from gdeep.utility import DEVICE


def test_gfdbc_2_dim():

    circle_detect_nn = FFNet((2, 2)).to(DEVICE)

    g = GradientFlowDecisionBoundaryCalculator(
        model=circle_detect_nn,
        initial_points=torch.rand((100, 2)),
        optimizer=lambda params: torch.optim.Adam(params),
    )
    for i in range(100):
        g.step()
    assert g.get_decision_boundary().size() == torch.Size([100, 2])


def test_gfdbc_multiclass():

    circle_nn_3d = FFNet((3, 3)).to(DEVICE)

    g = GradientFlowDecisionBoundaryCalculator(
        model=circle_nn_3d,
        initial_points=torch.rand((100, 3)),
        optimizer=lambda params: torch.optim.Adam(params),
    )
    for i in range(100):
        g.step()
    assert g.get_decision_boundary().size() == torch.Size([100, 3])


# TODO: Check if significant number of points lie close to the decision boundary



from collections import Counter, OrderedDict
from typing import Callable, Tuple, Union, Any, Optional, List, Dict
from functools import partial as Partial  # noqa

import torch
from torch.utils.data import Dataset
from torchtext.data.utils import get_tokenizer
from torchtext.vocab import vocab
from torchtext.vocab.vocab import Vocab

from gdeep.utility import DEVICE
from ..abstract_preprocessing import AbstractPreprocessing
from .._utils import MissingVocabularyError

# type definition
from gdeep.utility.custom_types import Tensor


class TokenizerTextClassification(
    AbstractPreprocessing[Tuple[Any, str], Tuple[Tensor, Tensor]]
):
    """Preprocessing class. This class is useful to convert the data format
    ``(label, text)`` into the proper tensor format ``( word_embedding, label)``.
    The labels shouzld be interagers; f they ar string, they will be converted.

    Args:
        tokenizer :
            the tokenizer of the source text
        vocabulary :
            the vocubulary; it can be built of it can be
            given.

    """

    max_length: int
    is_fitted: bool
    vocabulary: Optional[Vocab]
    tokenizer: Optional[Partial]
    counter: Dict[str, int]
    counter_label: Dict[str, int]

    def __init__(
        self, tokenizer: Optional[Partial] = None, vocabulary: Optional[Vocab] = None
    ):
        if tokenizer is None:
            self.tokenizer = get_tokenizer("basic_english")
        else:
            self.tokenizer = tokenizer

        self.vocabulary = vocabulary
        self.max_length = 0
        self.is_fitted = False

    def fit_to_dataset(self, dataset: Dataset[Tuple[Any, str]]) -> None:
        """Method to extract global data, like to length of
        the sentences to be able to pad.

        Args:
            dataset :
                the data in the format ``(label, text)``

        """

        self.counter = Counter()  # for the text
        self.counter_label = Counter()  # for the text
        self.ordered_dict = OrderedDict()
        self.list_of_possible_labels: List[Any] = []
        for (label, text) in dataset:  # type: ignore
            # if isinstance(text, tuple) or isinstance(text, list):
            #    text = text[0]
            self.counter.update(self.tokenizer(text))  # type: ignore
            self.counter_label.update([label])
            self.max_length = max(self.max_length, len(self.tokenizer(text)))  # type: ignore
        # build the vocabulary
        if not self.vocabulary:
            self.ordered_dict = OrderedDict(
                sorted(self.counter.items(), key=lambda x: x[1], reverse=True)
            )
            self.list_of_possible_labels = list(self.counter_label.keys())
            self.vocabulary = vocab(self.ordered_dict)
            unk_token = "<unk>"  # type: ignore
            if unk_token not in self.vocabulary:
                self.vocabulary.insert_token(unk_token, 0)
            self.vocabulary.set_default_index(self.vocabulary[unk_token])
        self.is_fitted = True
        # self.save_pretrained(".")

    def __call__(self, datum: Tuple[Any, str]) -> Tuple[Tensor, Tensor]:
        """This method is applied to each datum and
        transforms it following the rule below

        Args:
            datum (tuple):
                a single datum, being it a tuple
                with ``(label, text)``
        """
        # if not self.is_fitted:
        #    self.load_pretrained(".")
        if self.vocabulary:
            text_pipeline: Callable[[str], List[int]] = lambda x: [
                self.vocabulary[token] for token in self.tokenizer(x)  # type: ignore
            ]  # type: ignore
        else:
            raise MissingVocabularyError(
                "Please fit this preprocessor to initialise the vocabulary"
            )
        pad_item: int = 0

        _text = datum[1]
        # if isinstance(_text, tuple) or isinstance(_text, list):
        #    _text = _text[1]
        processed_text = torch.tensor(text_pipeline(_text), dtype=torch.long).to(DEVICE)
        # convert to tensors (padded)
        out_text = torch.cat(
            [
                processed_text,
                pad_item
                * torch.ones(self.max_length - processed_text.shape[0]).to(DEVICE),
            ]
        ).to(torch.long)
        # preprocess labels
        label_pipeline = lambda x: torch.tensor(x, dtype=torch.long)

        _label = datum[0]
        out_label = label_pipeline(self.list_of_possible_labels.index(_label)).to(
            DEVICE
        )
        return out_text, out_label



from .attention_pooling_layer import AttentionPoolingLayer
from .max_pooling_layer import MaxPoolingLayer
from .mean_pooling_layer import MeanPoolingLayer
from .sum_pooling_layer import SumPoolingLayer

__all__ = [
    "AttentionPoolingLayer",
    "MaxPoolingLayer",
    "MeanPoolingLayer",
    "SumPoolingLayer",
]



from typing import Optional

import torch
import torch.nn as nn
from torch.nn import Module, MultiheadAttention

from ..persformer_config import PersformerConfig

# Type aliases
from gdeep.utility.custom_types import Tensor


class AttentionPoolingLayer(Module):
    """This class implements the attention mechanism
    with a pooling layer to enforce permutation
    invariance"""

    config: PersformerConfig

    def __init__(self, config: PersformerConfig):
        super().__init__()
        self.config = config
        self.query = nn.parameter.Parameter(
            torch.Tensor(1, config.hidden_size), requires_grad=True
        )
        self.scaled_dot_product_attention = MultiheadAttention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_probs_dropout_prob,
            batch_first=True,
        )

    def forward(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, hidden_size)

        Returns:
            The pooled output. Of shape (batch_size, hidden_size)
        """
        output, _ = self.scaled_dot_product_attention(
            self.query.expand(input_batch.shape[0], -1, -1),
            input_batch,
            input_batch,
            key_padding_mask=attention_mask,
        )
        return output.squeeze(dim=1)



from typing import List, Sequence, Tuple, Union
import numpy as np  # type: ignore
import multiprocessing
import torch  # type: ignore
from torch.utils.data import DataLoader, TensorDataset  # type: ignore
from sklearn.model_selection import train_test_split  # type: ignore
from einops import rearrange  # type: ignore
from gtda.homology import WeakAlphaPersistence  # type: ignore
from sympy import N, S


class DataLoaderKwargs(object):
    """Object to store keyword arguments for train, val, and test dataloaders"""

    def __init__(self, *, train_kwargs, val_kwargs, test_kwargs):
        self.train_kwargs = train_kwargs
        self.val_kwargs = val_kwargs
        self.test_kwargs = test_kwargs

    def get_train_args(self):
        return self.train_kwargs

    def get_val_args(self):
        return self.val_kwargs

    def get_test_args(self):
        return self.test_kwargs


class OrbitsGenerator(object):
    """Generate Orbit dataset consistent of orbits defined by the dynamical system
    x[n+1] = x[n] + r * y[n] * (1  - y[n]) % 1
    y[n+1] = y[n] + r * x[n+1] * (1 - x[n+1]) % 1
    Note that there is an x[n+1] value in the second dimension.
    The parameter r is an hyperparameter and the classification task is to predict
    it given the orbit.
    By default r is chosen from (2.5, 3.5, 4.0, 4.1, 4.3).
    Args:
        parameters (Tuple[float]):
            Hyperparameter of the dynamical systems.
        num_orbits_per_class (int):
            number of orbits per class.
        num_pts_per_orbit (int):
            number of points per orbit.
        homology_dimensions (Sequence[int]):
            homology dimension of the persistence
            diagrams.
        validation_percentage (float, optional):
            Percentage of the validation dataset.
            Defaults to 0.0.
        test_percentage (float, optional):
            Percentage of the test dataset. Defaults to 0.0.
        dynamical_system (str, optional):
            either use persistence paths
            convention ´pp_convention´ or the classical convention
            ´classical_convention´. Defaults to '´classical_convention´'.
        n_jobs (int, optional):
            number of cpus to run the computation on. Defaults to 1.
    """

    def __init__(
        self,
        parameters: Sequence[float] = (2.5, 3.5, 4.0, 4.1, 4.3),
        num_orbits_per_class: int = 1_000,
        num_pts_per_orbit: int = 1_000,
        homology_dimensions: Sequence[int] = (0, 1),
        validation_percentage: float = 0.0,
        test_percentage: float = 0.0,
        dynamical_system: str = "classical_convention",
        n_jobs: int = 1,
        dtype: str = "float32",
        arbitrary_precision=False,
    ) -> None:

        # Initialize member variables.
        self._parameters = parameters
        self._num_classes = len(self._parameters)
        self._num_orbits_per_class = num_orbits_per_class
        self._num_pts_per_orbit = num_pts_per_orbit
        self._homology_dimensions = homology_dimensions
        self._num_homology_dimensions = len(self._homology_dimensions)
        self._n_jobs = n_jobs
        # Assert that validation and testing parameters are valid
        assert (
            test_percentage >= 0.0
            and validation_percentage >= 0.0
            and test_percentage + validation_percentage < 1.0
        )
        self._validation_percentage = validation_percentage
        self._test_percentage = test_percentage

        # Assert that convention for the dynamical system is valid
        assert dynamical_system in (
            "classical_convention",
            "pp_convention",
        ), f"{dynamical_system} is not supported"
        self.dynamical_system = dynamical_system

        # Initialize orbits array and persistence diagrams array with None.
        self._orbits = None
        self._labels = None
        self._persistence_diagrams = None

        # Initialize the train, val, and test indices with None.
        self._train_idcs = None
        self._val_idcs = None
        self._test_idcs = None

        assert dtype in (
            "float32",
            "float64",
            "float128",
        ), f"Type {dtype} is not supported."
        self._dtype = dtype

        self.arbitrary_precision = arbitrary_precision

    def orbits_from_array(self, orbits):
        assert (
            orbits.shape[0] == self._num_orbits_per_class * self._num_classes
            and orbits.shape[1] == self._num_pts_per_orbit
            and orbits.shape[2] == 2
        ), "Array does not have the right shape."

        self._orbits = orbits

        y = np.array(
            [self._num_orbits_per_class * [c] for c in range(self._num_classes)]
        )

        self._labels = y.reshape(-1)

    def _generate_orbits(self) -> None:
        """Generate the orbits for the dynamical system."""

        # If orbits are already computed do nothing
        if self._orbits is not None:
            return
        else:
            # Initialize the orbits array with zeros.
            x = np.zeros(
                (
                    self._num_classes,  # type: ignore
                    self._num_orbits_per_class,
                    self._num_pts_per_orbit,
                    2,
                )
            )
            # Initialize the labels array with the hyperparameter indices.
            y = np.array(
                [self._num_orbits_per_class * [c] for c in range(self._num_classes)]
            )

            self._labels = y.reshape(-1).astype("int64")  # type: ignore
            # generate dataset
            for class_idx, p in enumerate(self._parameters):  # type: ignore
                x[class_idx, :, 0, :] = np.random.rand(self._num_orbits_per_class, 2)  # type: ignore

                if self.arbitrary_precision:
                    assert (
                        self.dynamical_system == "classical_convention"
                    ), "Only classical_convention implemented yet"
                    for orbit in range(self._num_orbits_per_class):
                        print(orbit)
                        x[class_idx, orbit, :, :] = self._orbit_high_precision(
                            x_init=x[class_idx, orbit, 0, :],
                            rho=p,
                            num_points=self._num_pts_per_orbit,
                            precision=600,
                        )

                else:

                    for i in range(1, self._num_pts_per_orbit):  # type: ignore
                        x_cur = x[class_idx, :, i - 1, 0]
                        y_cur = x[class_idx, :, i - 1, 1]

                        if self.dynamical_system == "pp_convention":
                            x[class_idx, :, i, 0] = (
                                x_cur + p * y_cur * (1.0 - y_cur)
                            ) % 1
                            x[class_idx, :, i, 1] = (
                                y_cur + p * x_cur * (1.0 - x_cur)
                            ) % 1
                        else:
                            x[class_idx, :, i, 0] = (
                                x_cur + p * y_cur * (1.0 - y_cur)
                            ) % 1
                            x_next = x[class_idx, :, i, 0]
                            x[class_idx, :, i, 1] = (
                                y_cur + p * x_next * (1.0 - x_next)
                            ) % 1

            self._orbits = x.reshape((-1, self._num_pts_per_orbit, 2))  # type: ignore

    def _orbit_high_precision(self, x_init, rho, num_points=1_000, precision=600):
        x_precise = np.zeros((1_000, 2))

        x0 = S(x_init[0])
        y0 = S(x_init[1])

        for i in range(0, num_points):
            x_precise[i, 0] = x0
            x_precise[i, 1] = y0

            x0 = N((x0 + rho * y0 * (1.0 - y0)) % 1, precision)
            y0 = N((y0 + rho * x0 * (1.0 - x0)) % 1, precision)

        return x_precise

    def _compute_persistence_diagrams(self) -> None:
        """Computes the weak alpha persistence of the orbit data clouds.
        The result is stored in member variable ´persistence_diagrams´.
        It has the shape
        [num_classes, num_orbits, num_persistence_points, 3].
        In the last dimension the first two values are the coordinates of
        the points in the persistence diagrams and the third is the
        homology dimension.
        """
        if self._orbits is None:
            self._generate_orbits()

        # compute weak alpha persistence
        wap = WeakAlphaPersistence(
            homology_dimensions=self._homology_dimensions, n_jobs=self._n_jobs
        )
        # c: class, o: orbit, p: point, d: dimension
        # orbits_stack = rearrange(self._orbits, 'c o p d -> (c o) p d')  # stack classes
        persistence_diagrams_categorical = wap.fit_transform(self._orbits)
        # shape: (num_classes * n_samples, n_features, 3)
        # Convert persistence diagram to one-hot homological dimension encoding
        self._persistence_diagrams = self._persistence_diagrams_to_one_hot(
            persistence_diagrams_categorical
        ).astype(  # type: ignore
            self._dtype
        )  # type: ignore

    def _persistence_diagrams_to_one_hot(self, persistence_diagrams):
        """Convert homology dimension to one-hot encoding
        Args:
            persistence_diagrams ([np.array]):
                persistence diagram with categorical
                homology dimension.
        Returns:
            [np.array]:
                persistent diagram with one-hot encoded homology dimension.
        """
        if self._num_homology_dimensions > 1:
            return np.concatenate(
                (
                    persistence_diagrams[:, :, :2],  # point coordinates
                    (
                        np.eye(self._num_homology_dimensions)[  # type: ignore
                            persistence_diagrams[:, :, -1].astype(np.int32)
                        ]
                    ),
                ),
                axis=-1,
            )
        else:
            return persistence_diagrams[:, :, :2]

    def get_orbits(self) -> Union[None, np.ndarray]:
        """Returns the orbits as an ndarrays of shape
        (num_classes * num_orbits_per_class, num_pts_per_orbit, 2)
        Returns:
            np.ndarray:
                Orbits
        """
        if self._orbits is None:
            self._generate_orbits()
        return self._orbits

    def get_persistence_diagrams(self) -> Union[None, np.ndarray]:
        """Returns the orbits as an ndarrays of shape
        (num_classes * num_orbits_per_class, num_topological_features, 3)
        Returns:
            np.ndarray:
                Persistence diagrams
        """
        if self._persistence_diagrams is None:
            self._compute_persistence_diagrams()

        return self._persistence_diagrams

    def _split_data_idcs(self) -> None:
        """Split the data indices into training, validation and testing data."""
        idcs = np.arange(self._num_classes * self._num_orbits_per_class)

        if self._test_percentage > 0.0:
            rest_idcs, self._test_idcs = train_test_split(
                idcs, test_size=self._test_percentage
            )
        else:
            rest_idcs, self._test_idcs = idcs, []  # type: ignore

        if self._validation_percentage > 0.0:
            self._train_idcs, self._val_idcs = train_test_split(
                rest_idcs,
                test_size=(self._validation_percentage / (1.0 - self._test_percentage)),
            )
        else:
            self._train_idcs, self._val_idcs = rest_idcs, []  # type: ignore

    def _get_data_loaders(
        self, list_of_arrays: List[np.ndarray], dataloaders_kwargs: DataLoaderKwargs
    ) -> Tuple[DataLoader, DataLoader, DataLoader]:
        """Generates a DataLoader for the given list of arrays.
        Args:
            list_of_arrays ([List[np.ndarray]]):
                List of arrays to load.

        Returns:
            Tuple[DataLoader, DataLoader, DataLoader]:
                train, val, test data loaders.
        """
        assert (
            (self._train_idcs is not None)
            and (self._val_idcs is not None)
            and (self._test_idcs is not None)
        ), "Train, validation, and test data must be initialized"

        dl_train = DataLoader(
            TensorDataset(*(torch.tensor(a[self._train_idcs]) for a in list_of_arrays)),
            **dataloaders_kwargs.get_train_args(),
        )
        dl_val = DataLoader(
            TensorDataset(*(torch.tensor(a[self._val_idcs]) for a in list_of_arrays)),
            **dataloaders_kwargs.get_val_args(),
        )
        dl_test = DataLoader(
            TensorDataset(*(torch.tensor(a[self._train_idcs]) for a in list_of_arrays)),
            **dataloaders_kwargs.get_test_args(),
        )
        return dl_train, dl_val, dl_test

    def get_dataloader_orbits(
        self, dataloaders_kwargs: DataLoaderKwargs
    ) -> Tuple[DataLoader, DataLoader, DataLoader]:
        """Generates a Dataloader from the orbits dataset
        Returns:
            DataLoader:
                Dataloader of orbits
        """
        if self._orbits is None:
            self._generate_orbits()
        if self._train_idcs is None:
            self._split_data_idcs()
        return self._get_data_loaders(
            [
                self._orbits.astype(self._dtype),  # type: ignore
                self._labels,  # type: ignore
            ],  # type: ignore
            dataloaders_kwargs,
        )

    def get_dataloader_persistence_diagrams(
        self, dataloaders_kwargs: DataLoaderKwargs
    ) -> Tuple[DataLoader, DataLoader, DataLoader]:
        """Generates a Dataloader from the persistence diagrams dataset
        Returns:
            Tuple[DataLoader, DataLoader, DataLoader]:
                Dataloaders of persistence diagrams
        """
        if self._persistence_diagrams is None:
            self._compute_persistence_diagrams()
        if self._train_idcs is None:
            self._split_data_idcs()

        return self._get_data_loaders(
            [self._persistence_diagrams, self._labels],  # type: ignore
            dataloaders_kwargs,
        )

    def get_dataloader_combined(
        self,
        dataloaders_kwargs: DataLoaderKwargs,
    ) -> Tuple[DataLoader, DataLoader, DataLoader]:
        """Generates a Dataloader from the orbits dataset and the persistence diagrams
        Returns:
            DataLoader:
                Dataloader of orbits and persistence diagrams
        """
        if self._persistence_diagrams is None:
            self._compute_persistence_diagrams()
        if self._train_idcs is None:
            self._split_data_idcs()
        return self._get_data_loaders(
            [
                self._orbits,  # type: ignore
                self._persistence_diagrams,  # type: ignore
                self._labels,  # type: ignore
            ],  # type: ignore
            dataloaders_kwargs,
        )


def generate_orbit_parallel(
    num_classes,
    num_orbits,
    num_pts_per_orbit: int = 100,
    parameters: List[float] = [1.0],
) -> np.ndarray:
    """Generate sequence of points of a dynamical system
    in a parallel manner.

    Args:
        num_classes (int):
            number of classes of dynamical systems.
        num_orbits (int):
            number of orbits of dynamical system per class.
        num_pts_per_orbit (int, optional):
            Number of points to generate.
            Defaults to 100.
        parameter (List[float], optional):
            List of parameters of the dynamical
            system.
            Defaults to [1.0].

    Returns:
        np.ndarray:
            Array of sampled points of the dynamical system.
    """
    try:
        for parameter in parameters:
            assert parameter > 0
    except AssertionError:
        print("Parameter must be greater than 0")
    try:
        assert num_pts_per_orbit > 0
    except AssertionError:
        print("num_pts_per_orbit must be greater than 0")

    x = np.zeros((num_classes, num_orbits, num_pts_per_orbit, 2))

    # generate dataset
    for cidx, p in enumerate(parameters):  # type: ignore
        x[cidx, :, 0, :] = np.random.rand(num_orbits, 2)  # type: ignore

        for i in range(1, num_pts_per_orbit):  # type: ignore
            x_cur = x[cidx, :, i - 1, 0]
            y_cur = x[cidx, :, i - 1, 1]

            x[cidx, :, i, 0] = (x_cur + p * y_cur * (1.0 - y_cur)) % 1
            x_next = x[cidx, :, i, 0]
            x[cidx, :, i, 1] = (y_cur + p * x_next * (1.0 - x_next)) % 1

    assert not np.allclose(x[0, 0], x[0, 1])
    return x


def create_pd_orbits(
    orbits,
    num_classes,
    homology_dimensions=(0, 1),
    n_jobs=multiprocessing.cpu_count(),
) -> torch.Tensor:
    """Computes the weak alpha persistence of the orbit data clouds.

    Args:
        orbits (np.array):
            Orbits of shape [n_points, 2]
        homology_dimensions (tuple, optional):
            Dimensions to compute the
            persistence diagrams.
            Defaults to (0, 1).
        n_jobs (int, optional):
            Number of cpus to use for parallel computation.
            Defaults to multiprocessing.cpu_count().

    Returns:
        np.array:
            Array of persistence diagrams of shape
            [num_classes, num_orbits, num_persistence_points, 3].
            In the last dimension the first two values are the coordinates of
            the points in the persistence diagrams and the third is the
            homology dimension.
    """
    # compute weak alpha persistence
    wap = WeakAlphaPersistence(homology_dimensions=homology_dimensions, n_jobs=n_jobs)
    # c: class, o: orbit, p: point, d: dimension
    orbits_stack = rearrange(orbits, "c o p d -> (c o) p d")  # stack classes
    diagrams = wap.fit_transform(orbits_stack)
    # shape: (num_classes * n_samples, n_features, 3)

    # combine class and orbit dimensions
    diagrams = rearrange(
        diagrams, "(c o) p d -> c o p d", c=num_classes  # type: ignore
    )
    return diagrams


def convert_pd_orbits_to_tensor(
    diagrams: np.ndarray,
    num_classes: int,
    num_orbits: int,
    num_homology_dimensions: int = 2,
):
    """[summary]

    Args:
        diagrams (np.ndarray):
            [description]
        num_classes (int):
            [description]
        num_orbits (int):
            [description]
        num_homology_dimensions (int, optional):
            [description]. Defaults to 2.

    Returns:
        torch.tensor:
            tuple of
            1.
            2. label tensor.
    """
    # c: class, o: orbit, p: point in persistence diagram,
    # d: coordinates + homology dimension
    x = rearrange(diagrams, "c o p d -> (c o) p d", c=num_classes)  # type: ignore
    # convert homology dimension to one-hot encoding
    x = np.concatenate(
        (
            x[:, :, :2],
            (
                np.eye(num_homology_dimensions)[  # type: ignore
                    x[:, :, -1].astype(np.int32)
                ]
            ),
        ),
        axis=-1,
    )
    # convert from [orbit, sequence_length, feature] to
    # [orbit, feature, sequence_length] to fit to the
    # input_shape of `SmallSetTransformer`
    # x = rearrange(x, 'o s f -> o f s')

    # generate labels
    y_list = []
    for i in range(num_classes):  # type: ignore
        y_list += [i] * num_orbits  # type: ignore

    y = np.array(y_list)

    # load dataset to PyTorch dataloader

    x_tensor = torch.Tensor(x)
    y_tensor = torch.Tensor(y).long()

    return x_tensor, y_tensor



from re import I
from ._version import __version__

from . import (
    analysis,
    data,
    models,
    trainer,
    search,
    topology_layers,
    utility,
    visualization,
)

__all__ = [
    "analysis",
    "data",
    "models",
    "trainer",
    "search",
    "topology_layers",
    "utility",
    "visualization",
]



from gdeep.data.datasets import DatasetCloud, dataset_cloud

import hashlib
import logging
import os
from os import remove, environ
from os.path import join, exists
from shutil import rmtree

import numpy as np  # type: ignore
import torch

from gdeep.utility.utils import get_checksum

LOGGER = logging.getLogger(__name__)


# Test public access for downloading datasets
def test_public_access():
    # Download a small dataset from Google Cloud Storage
    dataset = "SmallDataset"
    download_directory = join("examples", "data", "DatasetCloud", "Tmp")

    # Remove download directory recursively if it exists
    if exists(download_directory):
        rmtree(download_directory)

    # Create download directory
    os.makedirs(download_directory, exist_ok=False)

    dataset_cloud = DatasetCloud(
        dataset, download_directory=download_directory, use_public_access=True
    )
    dataset_cloud.download()

    # Check if the downloaded files (metadata.json, data.json, labels.json)
    # are correct
    checksums = {
        "data.pt": "2ef68a718e29134cbcbf46c9592f6168",
        "labels.pt": "d71992425033c6bf449d175db146a423",
    }

    for file in checksums.keys():
        assert (
            get_checksum(join(download_directory, dataset, file)) == checksums[file]
        ), "File {} is corrupted.".format(file)
    # Recursively remove download directory
    rmtree(download_directory)


def test_get_dataset_list():
    # Download directory will not be used as well ass the dataset
    # It's only used for initialization of the DatasetCloud object
    download_directory = ""
    dataset_cloud = DatasetCloud(
        "SmallDataset", download_directory=download_directory, use_public_access=True
    )
    dataset_list = dataset_cloud.get_existing_datasets()
    assert len(dataset_list) > 0, "Dataset list is empty."
    assert "SmallDataset" in dataset_list, "Dataset list does not contain the dataset."

    # Test if the list does not contain duplicates
    assert len(dataset_list) == len(
        set(dataset_list)
    ), "Dataset list contains duplicates."


if "GOOGLE_APPLICATION_CREDENTIALS" in dict(environ):

    def test_update_dataset_list():
        # Create DatasetCloud object
        dataset_cloud = DatasetCloud("", use_public_access=False)
        # Update the dataset list
        dataset_cloud._update_dataset_list()

    def test_upload_and_download():
        for data_format in ["pytorch_tensor", "numpy_array"]:
            download_directory = join("examples", "data", "DatasetCloud")
            # Generate a dataset
            # You don't have to do that if you already have a pickled dataset
            size_dataset = 100
            input_dim = 5
            num_labels = 2

            if data_format == "pytorch_tensor":
                data = torch.rand(size_dataset, input_dim)
                labels = torch.randint(0, num_labels, (size_dataset,)).long()

                # pickle data and labels
                data_filename = "tmp_data.pt"
                labels_filename = "tmp_labels.pt"
                torch.save(data, data_filename)
                torch.save(labels, labels_filename)
            elif data_format == "numpy_array":
                data = np.random.rand(size_dataset, input_dim)  # type: ignore
                labels = np.random.randint(
                    0, num_labels, (size_dataset,), dtype=np.long
                )

                # pickle data and labels
                data_filename = "tmp_data.npy"
                labels_filename = "tmp_labels.npy"
                np.save(data_filename, data)
                np.save(labels_filename, labels)
            else:
                raise ValueError(f"Unknown data format: {data_format}")

            ## Upload dataset to Cloud
            dataset_name = "TmpSmallDataset"
            dataset_cloud = DatasetCloud(
                dataset_name,
                download_directory=download_directory,
                use_public_access=False,
            )

            # Specify the metadata of the dataset
            dataset_cloud._add_metadata(
                name=dataset_name,
                size_dataset=size_dataset,
                input_size=(input_dim,),
                num_labels=num_labels,
                data_type="tabular",
                data_format=data_format,
            )

            # upload dataset to Cloud
            dataset_cloud._upload(data_filename, labels_filename)

            # download dataset from Cloud to ´example/data/DataCloud/SmallDataset/´
            dataset_cloud.download()

            # remove created blob
            dataset_cloud._data_cloud.delete_blobs(dataset_name)

            # check whether downloaded dataset is the same as the original dataset
            if data_format == "pytorch_tensor":
                downloaded_files = ["data.pt", "labels.pt", "metadata.json"]
            elif data_format == "numpy_array":
                downloaded_files = ["data.npy", "labels.npy", "metadata.json"]
            else:
                raise ValueError(f"Unknown data format: {data_format}")
            for file in downloaded_files:
                hash_original = get_checksum("tmp_" + file)
                path_downloaded_file = join(download_directory, dataset_name, file)
                hash_downloaded = get_checksum(path_downloaded_file)
                assert (
                    hash_original == hash_downloaded
                ), "Original and downloaded files do not match."

            # remove the labels and data files
            remove(data_filename)
            remove(labels_filename)

            # remove the downloaded dataset
            rmtree(join(download_directory, dataset_name))

            # remove the metadata file
            # will get deleted automatically when dataset_cloud is out of scope.
            del dataset_cloud



from typing import Tuple

import gudhi as gd  # type: ignore
import numpy as np
from scipy.linalg import eigh
from scipy.sparse import csgraph

from gdeep.utility.utils import flatten_list_of_lists
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
    get_one_hot_encoded_persistence_diagram_from_gudhi_extended,
)

from gdeep.utility.custom_types import Array


# Compute the heat kernel signature of a graph.
def _heat_kernel_signature(adj_mat: Array, diffusion_parameter: float = 1.0) -> Array:
    """Given a graph adjacency matrix, compute the heat kernel signature.

    Args:
        adj_mat:
            The adjacency matrix of the graph.
        diffusion_parameter:
            The diffusion parameter of the heat kernel.

    Returns:
        ndarray:
            The heat kernel signature of the graph. The shape of the array is
            (num_vertices).
    """
    eigenvals, eigenvectors = _get_eigenvalues_eigenvectors(adj_mat)
    hks = (np.square(eigenvectors) * np.exp(-diffusion_parameter * eigenvals)).sum(  # type: ignore
        axis=1
    )
    return hks


def _get_eigenvalues_eigenvectors(adj_mat) -> Tuple[Array, Array]:
    """Compute the eigenvector and eigenvalue of the adjacency matrix.

    Args:
        adj_mat:
            The adjacency matrix of the graph.

    Returns:
        (ndarray, ndarray):
            The eigenvectors and eigenvalues of the adjacency matrix.
    """
    # Compute the Laplacian matrix
    laplacian_mat = csgraph.laplacian(adj_mat, normed=True)
    # Compute the eigenvalues and eigenvectors of the Laplacian matrix
    eigenvalues, eigenvectors = eigh(laplacian_mat)
    return eigenvalues, eigenvectors


def graph_extended_persistence_hks(
    adj_mat: Array, diffusion_parameter: float = 1.0
) -> OneHotEncodedPersistenceDiagram:
    """Compute the extended persistence of a graph.

    Args:
        adj_mat:
            The adjacency matrix of the graph.
        diffusion_parameter:
            The diffusion parameter of the heat kernel.

    Returns:
        ndarray:
            The extended persistence of the graph.
    """
    # Compute the heat kernel signature
    hks = _heat_kernel_signature(adj_mat, diffusion_parameter)
    # Compute the extended persistence
    persistence_diagram = graph_extended_persistence_gudhi(adj_mat, hks)
    return get_one_hot_encoded_persistence_diagram_from_gudhi_extended(
        persistence_diagram
    )


def graph_extended_persistence_gudhi(
    A: Array, filtration_val: Array
) -> Tuple[Array, Array, Array, Array]:

    # TODO: Rewrite this function
    num_vertices = A.shape[0]
    (xs, ys) = np.where(np.triu(A))  # type: ignore
    st = gd.SimplexTree()  # type: ignore
    for i in range(num_vertices):
        st.insert([i], filtration=-1e10)
    for idx, x in enumerate(xs):
        st.insert([x, ys[idx]], filtration=-1e10)  # type: ignore
    for i in range(num_vertices):
        st.assign_filtration([i], filtration_val[i])
    st.make_filtration_non_decreasing()
    st.extend_filtration()
    extended_persistence = st.extended_persistence()
    dgmOrd0, dgmRel1, dgmExt0, dgmExt1 = (
        extended_persistence[0],
        extended_persistence[1],
        extended_persistence[2],
        extended_persistence[3],
    )
    dgmOrd0 = (
        np.vstack(
            [
                np.array([[min(p[1][0], p[1][1]), max(p[1][0], p[1][1])]])
                for p in dgmOrd0
                if p[0] == 0
            ]
        )
        if len(dgmOrd0)
        else np.empty([0, 2])
    )
    dgmRel1 = (
        np.vstack(
            [
                np.array([[min(p[1][0], p[1][1]), max(p[1][0], p[1][1])]])
                for p in dgmRel1
                if p[0] == 1
            ]
        )
        if len(dgmRel1)
        else np.empty([0, 2])
    )
    dgmExt0 = (
        np.vstack(
            [
                np.array([[min(p[1][0], p[1][1]), max(p[1][0], p[1][1])]])
                for p in dgmExt0
                if p[0] == 0
            ]
        )
        if len(dgmExt0)
        else np.empty([0, 2])
    )
    dgmExt1 = (
        np.vstack(
            [
                np.array([[min(p[1][0], p[1][1]), max(p[1][0], p[1][1])]])
                for p in dgmExt1
                if p[0] == 1
            ]
        )
        if len(dgmExt1)
        else np.empty([0, 2])
    )
    return dgmOrd0, dgmExt0, dgmRel1, dgmExt1  # type: ignore
